import pandas as pd
from usdm_excel.study_identifiers_sheet.study_references_sheet import StudyReferencesSheet
from usdm_model.code import Code

DEFAULT_COLUMNS = ['identifierScheme', 'organisationIdentifier', 'name', 'label', 'type', 'studyIdentifier', 'address', 'referenceType']
MISSING_COLUMN =  ['identifierScheme', 'organisationIdentifier', 'name', 'label', 'type', 'studyIdentifier', 'address']

def test_create(mocker, globals):
  org_1 = ('{"id": "Org_1", "name": "ClinicalTrials.gov", "label": "CT.gov Study Registry", '
             '"type": {"id": "Code_1", "code": "C93453", "codeSystem": "http://www.cdisc.org", "codeSystemVersion": "2023-12-15", "decode": "Study Registry", "instanceType": "Code"}, '
             '"identifierScheme": "USGOV", "identifier": "CT-GOV", '
             '"legalAddress": {"id": "Addr_1", "text": "line, city, district, state, postal_code, Denmark", "lines": ["line"], "city": "city", "district": "district", "state": "state", "postalCode": "postal_code", '
               '"country": {"id": "Code_2", "code": "DNK", "codeSystem": "ISO 3166 1 alpha3", "codeSystemVersion": "2020-08", "decode": "Denmark", "instanceType": "Code"}, '
               '"instanceType": "Address"}, '
             '"managedSites": [], '
             '"instanceType": "Organization"}'
  )
  expected_1 = ('{"id": "RI_1", "text": "NCT12345678", "scopeId": "Org_1", '
                '"instanceType": "ReferenceIdentifier", '
                '"type": {"id": "Code_3", "code": "C99910x1", "codeSystem": "http://www.cdisc.org", "codeSystemVersion": "2023-12-15", "decode": "Pediatric Investigation Plan", "instanceType": "Code"}}'
  )
  org_2 = ('{"id": "Org_2", "name": "ClinicalTrials2.gov", "label": "CT.gov Registry", '
             '"type": {"id": "Code_4", "code": "C93453", "codeSystem": "http://www.cdisc.org", "codeSystemVersion": "2023-12-15", "decode": "Study Registry", "instanceType": "Code"}, '
             '"identifierScheme": "USGOV2", "identifier": "CT-GOV2", '
             '"legalAddress": {"id": "Addr_2", "text": "line2, city2, district2, state2, postal_code2, Denmark", "lines": ["line2"], "city": "city2", "district": "district2", "state": "state2", "postalCode": "postal_code2", '
               '"country": {"id": "Code_5", "code": "DNK", "codeSystem": "ISO 3166 1 alpha3", "codeSystemVersion": "2020-08", "decode": "Denmark", "instanceType": "Code"}, '
               '"instanceType": "Address"}, '
             '"managedSites": [], '
             '"instanceType": "Organization"}'
  )
  expected_2 = ('{"id": "RI_2", "text": "NCT12345679", "scopeId": "Org_2", '
                '"instanceType": "ReferenceIdentifier", '
                '"type": {"id": "Code_6", "code": "C142424", "codeSystem": "http://www.cdisc.org", "codeSystemVersion": "2023-12-15", "decode": "Clinical Development Plan", "instanceType": "Code"}}'
  )
  ids = ['Code_1', 'Addr_1', 'Code_2', 'Org_1', 'Code_3', 'RI_1', 'Code_4', 'Addr_2', 'Code_5', 'Org_2', 'Code_6', 'RI_2']
  data = [
    [ 'USGOV',  'CT-GOV',  'ClinicalTrials.gov',  'CT.gov Study Registry', 'Study Registry', 'NCT12345678', 'line|district|city|state|postal_code|GBR', 'Pediatric Investigation Plan'],
    [ 'USGOV2', 'CT-GOV2', 'ClinicalTrials2.gov', 'CT.gov Registry',       'Study Registry', 'NCT12345679', 'line2,district2,city2,state2,postal_code2,FRA', 'Clinical Development Plan']
  ]
  sheet = _setup(mocker, globals, data, ids)
  assert len(sheet.items) == 2
  assert sheet.items[0].to_json() == expected_1
  assert sheet.organizations[0].to_json() == org_1
  assert sheet.items[1].to_json() == expected_2
  assert sheet.organizations[1].to_json() == org_2
    
def test_create_empty(mocker, globals):
  ids = []
  data = []
  sheet = _setup(mocker, globals, data, ids)
  assert len(sheet.items) == 0

def test_error(mocker, globals):
  ids = ['Code_1', 'Addr_1', 'Code_2', 'Org_1', 'Code_3', 'RI_1']
  data = {'organisationIdentifierScheme': ['USGOV'], 'organisationIdentifier': ['CT-GOV'], 'organisationName': [''], 'organisationType': ['Study Registry'], 'studyIdentifier': ['NCT12345678'], 'organisationAddress': ['line|district|city|state|postal_code|GBR']}
  mock_error = mocker.patch("usdm_excel.errors_and_logging.errors.Errors.add")
  sheet = _setup(mocker, globals, data, ids, MISSING_COLUMN)
  assert sheet.items == []
  assert mock_error.call_count == 3
  errors = [
    mocker.call('studyReferences', 1, 5, 'Empty cell detected where CDISC CT value expected.', 40),
    mocker.call('studyReferences', 1, 7, "Address '' does not contain the required fields (lines, district, city, state, postal code and country code) using ',' separator characters, only 0 found", 40),
    mocker.call('studyReferences', None, None, 'Exception. Failed to create Organization object. See log for additional details.', 40)
  ]
  mock_error.assert_has_calls(errors)
  
def _setup(mocker, globals, data, ids, columns=DEFAULT_COLUMNS):
  globals.cross_references.clear()
  mock_present = mocker.patch("usdm_excel.base_sheet.BaseSheet._sheet_present")
  mock_present.side_effect=[True]
  mock_id = mocker.patch("usdm_excel.id_manager.IdManager.build_id")
  mock_id.side_effect = ids
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=columns)
  mock_json = mocker.patch("json.load")
  mock_json.return_value = {}
  return StudyReferencesSheet("", globals)



import pandas as pd
from usdm_excel.study_design_timing_sheet.study_design_timing_sheet import StudyDesignTimingSheet
from usdm_excel.study_design_timing_sheet.window_type import WindowType


def test_create(mocker, globals):
  globals.errors_and_logging.errors().clear()
  mock_present = mocker.patch("usdm_excel.base_sheet.BaseSheet._sheet_present")
  mock_present.side_effect=[True]
  mock_id = mocker.patch("usdm_excel.id_manager.IdManager.build_id")
  mock_id.side_effect=['Range1', 'Code1', 'Code2', 'TimingId_1', 'Code3', 'Code4', 'TimingId_2', 'Code5', 'Code6', 'TimingId_3']
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [
    ['Timing 1', 'Timing One', 'T1', 'After', 'X1', 'X2', '3 days', 'S2S', '-1..1 day'],
    ['Timing 2', 'Timing Two', 'T2', 'Before', 'X2', 'X3', '10 days', 'E2S', ''], 
    ['Timing 3', 'Timing Three', 'T3', 'Fixed', 'X4', 'X5', '3 days', 'e2e', '']
  ]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['name', 'description', 'label', 'type', 'from', 'to', 'timingValue', 'toFrom', 'window'])

  items = StudyDesignTimingSheet("", globals)
  #print(f"ERRORS: {[item.to_dict() for item in globals.errors_and_logging.errors().items]}")
  #print(f"ITEMS: {items.items}")
  assert len( globals.errors_and_logging.errors().items) == 0
  assert len(items.items) == 3
  assert items.items[0].id == 'TimingId_1'
  assert items.items[0].name == 'Timing 1'
  assert items.items[0].description == 'Timing One'
  assert items.items[0].type.decode == 'After'
  assert items.items[1].id == 'TimingId_2'
  assert items.items[1].description == 'Timing Two'
  assert items.items[1].type.decode == 'Before'
  assert items.items[2].id == 'TimingId_3'
  assert items.items[2].type.decode == 'Fixed Reference'

# def test_timepoint_type(mocker, globals):
#   mocked_open = mocker.mock_open(read_data="File")
#   mocker.patch("builtins.open", mocked_open)
#   mock_read = mocker.patch("pandas.read_excel")
#   data = { 'col_1': [ 'A:', 'P: 2D ', 'P2:2D', 'C:', 'N3: 3 weeks', 'N: +2 days', 'P: -3 hrs' ] }
#   mock_read.return_value = pd.DataFrame.from_dict(data)
#   parent = BaseSheet("", "")
#   test_data = [
#     (0,0,'anchor', 0, None, ISO8601Duration.ZERO_DURATION),
#     (1,0,'previous', -1, "2D", "P2D"),
#     (2,0,'previous', -2, "2D", "P2D"),
#     (3,0,'cycle start', 1, None, ISO8601Duration.ZERO_DURATION),
#     (4,0,'next', 3, "3 weeks", "P3W"),
#     (5,0,'next', 1, "+2 days", "P2D"),
#     (6,0,'previous', -1, "-3 hrs", "PT3H"),
#   ]
#   for index, test in enumerate(test_data):
#     item = TimepointType(parent, test[0], test[1])
#     assert(item.timing_type) == test[2]
#     assert(item.relative_ref) == test[3]
#     assert(item.description) == test[4]
#     assert(item.value) == test[5]

# def test_timepoint_type_error(mocker, globals):
#   mocked_open = mocker.mock_open(read_data="File")
#   mocker.patch("builtins.open", mocked_open)
#   mock_read = mocker.patch("pandas.read_excel")
#   data = { 'col_1': [ 'A', 'P: 2 ', '', 'P2: 2 decades' ] }
#   mock_read.return_value = pd.DataFrame.from_dict(data)
#   parent = BaseSheet("", "Sheet X")
#   test_data = [
#     (0,0,"Could not decode the timing value, no ':' detected in 'A'"),
#     (1,0,"Could not decode the duration value, no value and units found in '2'"),
#     (2,0,"Could not decode the timing value, cell was empty"),
#     (3,0,"Could not decode the duration value '2 decades'"),
#   ]
#   for index, test in enumerate(test_data):
#     mock_error = mocker.patch("usdm_excel.errors_and_logging.errors.Errors.add")
#     item = TimepointType(parent, test[0], test[1])
#     mock_error.assert_called()
#     assert mock_error.call_args[0][0] == "Sheet X"
#     assert mock_error.call_args[0][1] == test[0] + 1
#     assert mock_error.call_args[0][2] == test[1] + 1
#     assert mock_error.call_args[0][3] == test[2]

def test_window_type(mocker, globals):
  test_data = [
    ('1..1 Days', "P1D", "P1D", '1..1 Days'),
    (' -1..1 days', "P1D", "P1D", '-1..1 days'),
    ('-1 .. 1 weeks ', "P1W", "P1W", '-1 .. 1 weeks'),
    (None, None, None, ''),
  ]
  for index, test in enumerate(test_data):
    globals.errors_and_logging.errors().clear()
    item = WindowType(test[0], globals)
    assert(item.lower) == test[1]
    assert(item.upper) == test[2]
    assert(item.label) == test[3]
    assert(item.errors) == [] 

def test_window_type_error(mocker, globals):
  test_data = [
    ('1.. Days',"Could not decode the range value, possible typographical errors '1.. Days'"),
    ('-1.1 days',"Could not decode the range value, possible typographical errors '-1.1 days'"),
    ('-1 .. 1',"Could not decode the range value, possible typographical errors '-1 .. 1'"),
    (' .. 1 Weeks',"Could not decode the range value '.. 1 Weeks'")
  ]
  for index, test in enumerate(test_data):
    globals.errors_and_logging.errors().clear()
    item = WindowType(test[0], globals)
    assert item.errors == [test[1]]



import logging
from .error import Error

class Errors():

  CRITICAL = Error.CRITICAL
  WARNING = Error.WARNING
  ERROR = Error.ERROR
  DEBUG = Error.DEBUG
  INFO = Error.INFO

  def __init__(self):
    self.items = []
    self._logger = logging.getLogger(__name__)

  def clear(self):
    self.items = []

  def add(self, message: str, level: int=Error.ERROR) -> None:
    error = Error(message, level)
    self.items.append(error)      
    self._logger.log(level, error.to_log())

  def count(self) -> int:
    return len(self.items)
  
  def dump(self, level):
    result = []
    for item in self.items:
      if item.level >= level:
        result.append(item.to_dict())
    return result




from typing import Literal, Union
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .quantity import Quantity
from .range import Range

class Strength(ApiBaseModelWithIdNameLabelAndDesc):
  denominator: Union[Quantity, None] = None
  numerator: Union[Quantity, Range, None] = None
  instanceType: Literal['Strength']




from yattag import Doc
from usdm_model.eligibility_criterion import EligibilityCriterion
from usdm_model.narrative_content import NarrativeContent, NarrativeContentItem
from usdm_db.fhir.to_fhir import ToFHIR
from tests.test_factory import Factory
from tests.test_data_factory import MinimalStudy
from uuid import UUID
from bs4 import BeautifulSoup   

fake_uuid = UUID(f'00000000-0000-4000-8000-{1:012}', version=4)

def create_criteria(factory: Factory, minimal: MinimalStudy):
  INCLUSION = factory.cdisc_code('C25532', 'Inc')
  EXCLUSION = factory.cdisc_code('C25370', 'Exc')
  item_list = [
    {'name': 'IE1', 'label': '', 'description': '', 'text': 'Only perform at baseline', 
     'dictionaryId': None, 'category': INCLUSION, 'identifier': '01', 'nextId': None, 'previousId': None, 'contextId': None
    },
    {'name': 'IE2', 'label': '', 'description': '', 'text': '<p>Only perform on males</p>', 
    'dictionaryId': None, 'category': INCLUSION, 'identifier': '02', 'nextId': None, 'previousId': None, 'contextId': None
    },
  ]
  results = factory.set(EligibilityCriterion, item_list)
  for criterion in results:
    minimal.population.criterionIds.append(criterion.id)
  minimal.study.versions[0].criteria = results
  #minimal.study.documentedBy[0].templateName = 'document' # Fix the document template
  return results

def test_create(mocker, globals, minimal, factory):
  x = create_criteria(factory, minimal)
  fhir = ToFHIR(minimal.study, 'Sponsor')
  assert fhir is not None

def test_content_to_section(mocker, globals, minimal, factory):
  x = create_criteria(factory, minimal)
  fhir = ToFHIR(minimal.study, 'Sponsor')
  item = factory.item(NarrativeContentItem, {'name': "NCI1", 'text': 'Something here for the text'})
  content = factory.item(NarrativeContent, {'name': "C1", 'sectionNumber': '1.1.1', 'displaySectionNumber': True, 'sectionTitle': 'Section Title', 'displaySectionTitle': True, 'contentItemId': item.id, 'childIds': []})
  result = fhir._content_to_section(content, item.text)
  expected = '{"title": "Section Title", "code": {"text": "section1.1.1-section-title"}, "text": {"status": "generated", "div": "Something here for the text"}}'
  assert result.json() == expected

def test_format_section(mocker, globals, minimal, factory):
  x = create_criteria(factory, minimal)
  fhir = ToFHIR(minimal.study, 'Sponsor')
  assert fhir._format_section_title('A Section Heading') == 'a-section-heading'

def test_clean_section_number(mocker, globals, minimal, factory):
  x = create_criteria(factory, minimal)
  fhir = ToFHIR(minimal.study, 'Sponsor')
  assert fhir._clean_section_number('1.1') == '1.1'
  assert fhir._clean_section_number('1.1.') == '1.1'

# def test_add_section_heading(mocker, globals, minimal, factory):
#   x = create_criteria(factory, minimal)
#   fhir = ToFHIR(minimal.study, 'Sponsor')
#   item = factory.item(NarrativeContentItem, {'name': "NCI1", 'text': '<div xmlns="http://www.w3.org/1999/xhtml">Something here for the text</div>'})
#   content = factory.item(NarrativeContent, {'name': "C1", 'sectionNumber': '1.1.1', 'displaySectionNumber': True, 'sectionTitle': 'Section Title', 'displaySectionTitle': True, 'contentItemId': item.id, 'childIds': []})
#   div = BeautifulSoup(item.text, 'html.parser')
#   assert fhir._add_section_heading(content, div) == '<div xmlns="http://www.w3.org/1999/xhtml"><p>1.1.1 Section Title</p>Something here for the text</div>'

def test_remove_line_feeds(mocker, globals, minimal, factory):
  x = create_criteria(factory, minimal)
  fhir = ToFHIR(minimal.study, 'Sponsor')
  text = "<p>CNS imaging (CT scan or MRI of brain) compatible with AD within past 1 year.</p>\n<p>The following findings are incompatible with AD:</p>\n"  
  assert fhir._remove_line_feeds(text) == "<p>CNS imaging (CT scan or MRI of brain) compatible with AD within past 1 year.</p><p>The following findings are incompatible with AD:</p>"  



from usdm_excel.base_sheet import BaseSheet
from usdm_excel.study_soa_v2_sheet.soa_column_rows import SoAColumnRows
from usdm_excel.study_soa_v2_sheet.soa_activity import SoAActivity

class SoAActivities():
  
  def __init__(self, parent: BaseSheet):
    self._parent_sheet = parent
    self.items = []
    self._map = {}
    self._parent_activity = None
    for row_index, col_def in self._parent_sheet.sheet.iterrows():
      if row_index >= SoAColumnRows.FIRST_ACTIVITY_ROW:
        activity = SoAActivity(self._parent_sheet, row_index, self._map)
        the_activity = activity.activity
        if the_activity:
          self.items.append(activity)

  def group_and_link(self):
    activities = []
    biomedical_concepts = []
    biomedical_concept_surrogates = []
    for item in self.items:
      the_activity = item.activity
      activities.append(the_activity)
      biomedical_concept_surrogates += item.usdm_biomedical_concept_surrogates
      biomedical_concepts += item.usdm_biomedical_concepts
    self._parent_sheet.double_link(activities, 'previousId', 'nextId')
    return activities, biomedical_concepts, biomedical_concept_surrogates

  def set_parents(self):
    parents = any([x.is_parent for x in self.items])
    if parents:
      parent_activity = None
      for item in self.items:
        the_activity = item.activity
        if item.is_parent:
          parent_activity = the_activity
        elif parent_activity:
          parent_activity.childIds.append(the_activity.id)
        else:
          self._parent_sheet._general_error(f"Child activity with name '{the_activity.name}' does not have a parent specified")

      



from typing import Literal, Union
from .api_base_model import ApiBaseModelWithId
from .alias_code import AliasCode

class Quantity(ApiBaseModelWithId):
  value: float
  unit: Union[AliasCode, None] = None
  instanceType: Literal['Quantity']



import pytest
import pandas as pd
from usdm_excel.study_design_eligibility_criteria_sheet.study_design_eligibility_criteria_sheet import StudyDesignEligibilityCriteriaSheet
from usdm_model.code import Code


xfail = pytest.mark.xfail
def test_create(mocker, globals):
  mock_present = mocker.patch("usdm_excel.base_sheet.BaseSheet._sheet_present")
  mock_present.side_effect=[True]
  mock_id = mocker.patch("usdm_excel.id_manager.IdManager.build_id")
  mock_id.side_effect=['Code_1', 'EligibilityId_1', 'Code_2', 'EligibilityId_2', 'Code_3', 'EligibilityId_3']
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [
    ['Inclusion', '01', 'INC01', 'The study age criterion', 'Age critierion', 'Subjects should be between 18 and 45 years old', "dictionary"], 
    ['Inclusion', '02', 'INC01', 'The study abc criterion',   'ABC critierion', 'Subjects should have ABC', "dictionary"], 
    ['Exclusion', '01', 'EXC01', 'Exclude those with all fingers',   'Fingers critierion', 'Subjects should not have all fingers', "dictionary"], 
 ]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['category', 'identifier', 'name', 'description', 'label', 'text', 'dictionary'])
  items = StudyDesignEligibilityCriteriaSheet("", globals)
  assert len(items.items) == 3
  assert items.items[0].id == 'EligibilityId_1'
  assert items.items[0].category.decode == 'Inclusion Criteria'
  assert items.items[0].identifier == '01'
  assert items.items[0].name == 'INC01'
  assert items.items[0].description == 'The study age criterion'
  assert items.items[0].label == 'Age critierion'
  assert items.items[0].text == 'Subjects should be between 18 and 45 years old'
  assert items.items[1].category.decode == 'Inclusion Criteria'
  assert items.items[2].category.decode == 'Exclusion Criteria'
  
def test_create_empty(mocker, globals):
  mock_present = mocker.patch("usdm_excel.base_sheet.BaseSheet._sheet_present")
  mock_present.side_effect=[True]
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = []
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['category', 'identifier', 'name', 'description', 'label', 'text', 'dictionary'])
  items = StudyDesignEligibilityCriteriaSheet("", globals)
  assert len(items.items) == 0

def test_read_cell_by_name_error(mocker, globals):
  globals.cross_references.clear()
  call_parameters = []
  
  def my_add(sheet, row, column, message, level=10):
    call_parameters.append((sheet, row, column, message, level))
    return None

  mock_present = mocker.patch("usdm_excel.base_sheet.BaseSheet._sheet_present")
  mock_present.side_effect=[True]
  mock_error = mocker.patch("usdm_excel.errors_and_logging.errors.Errors.add", side_effect=my_add)
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [
    ['Inclusion', '01', 'The study age criterion', 'Age critierion', 'Subjects should be between 18 and 45 years old', "dictionary"]
  ]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['category', 'identifier', 'name', 'label', 'text', 'dictionary'])
  items = StudyDesignEligibilityCriteriaSheet("", globals)
  mock_error.assert_called()
  assert call_parameters == [
    ('studyDesignEligibilityCriteria', 1, -1, "Error attempting to read cell 'description'. Exception: Failed to detect column(s) 'description' in sheet", 40),
    ('studyDesignEligibilityCriteria', 1, 5, "Dictionary 'dictionary' not found", 30),
    ('studyDesignEligibilityCriteria', None, None, "Unable to find dictionary with name 'dictionary'", 40)
  ]
  



import pytest
import pandas as pd
from usdm_excel.assigned_person_sheet.assigned_person_sheet import AssignedPersonSheet
from usdm_model.organization import Organization
from usdm_excel.globals import Globals
from tests.test_factory import Factory

xfail = pytest.mark.xfail

def test_create(factory, mocker, globals):
  data = {
    'name': ['AP1', 'AP2', 'AP3'], 
    'description': ['Desc One', 'Desc Two', 'Desc Three'],
    'label': ['Lable 1', 'L2', 'L3'],
    'jobTitle': ['Title 1', 'Title 2', 'Title 3'],
    'organization': ['Sponsor 1', 'Sponsor 2', '']
  }
  expected_1 = '{"id": "AP_1", "name": "AP1", "label": "Lable 1", "description": "Desc One", "jobTitle": "Title 1", "organizationId": "O_1", "instanceType": "AssignedPerson"}'
  expected_2 = '{"id": "AP_2", "name": "AP2", "label": "L2", "description": "Desc Two", "jobTitle": "Title 2", "organizationId": "O_2", "instanceType": "AssignedPerson"}'
  expected_3 = '{"id": "AP_3", "name": "AP3", "label": "L3", "description": "Desc Three", "jobTitle": "Title 3", "organizationId": null, "instanceType": "AssignedPerson"}'
  mock_id = mocker.patch("usdm_excel.id_manager.IdManager.build_id")
  mock_id.side_effect=['C_1', 'C_2', 'C_3', 'O_1', 'O_2', 'O_3', 'AP_1', 'AP_2', 'AP_3', 'X_4', 'X_5']
  _setup(mocker, globals, data)
  _create_orgs(factory, globals)
  item = AssignedPersonSheet("", globals)
  assert len(item.items) == 3
  assert item.items[0].to_json() == expected_1
  assert item.items[1].to_json() == expected_2
  assert item.items[2].to_json() == expected_3

def test_create_empty(mocker, globals):
  data = {}
  _setup(mocker, globals, data)
  item = AssignedPersonSheet("", globals)
  assert len(item.items) == 0

def test_read_cell_by_name_error(mocker, globals):
  data = {
    'name': ['AP1'], 
    'description': ['Desc One'],
    'label': ['Lable 1'],
    'organization': ['']
  }
  mock_error = mocker.patch("usdm_excel.errors_and_logging.errors.Errors.add")
  mock_id = mocker.patch("usdm_excel.id_manager.IdManager.build_id")
  mock_id.side_effect=['Code_1', 'Abbreviation_1']
  _setup(mocker, globals, data)
  item = AssignedPersonSheet("", globals)
  assert mock_error.call_count == 1
  mock_error.assert_has_calls([mocker.call('people', 1, -1, "Error attempting to read cell 'jobTitle'. Exception: Failed to detect column(s) 'jobTitle' in sheet", 40)])
  
def _setup(mocker, globals, data):
  globals.cross_references.clear()
  mock_present = mocker.patch("usdm_excel.base_sheet.BaseSheet._sheet_present")
  mock_present.side_effect=[True]
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame.from_dict(data)

def _create_orgs(factory: Factory, globals: Globals):
  items = [
    {'name': 'Sponsor 1', 'type': factory.cdisc_code("C70793", "sponsor"), 'identifier': "123456789", 'identifierScheme': "DUNS", 'legalAddress': None},
    {'name': 'Sponsor 2', 'type': factory.cdisc_code("C70793", "sponsor"), 'identifier': "222222222", 'identifierScheme': "DUNS", 'legalAddress': None}, 
    {'name': 'Sponsor 3', 'type': factory.cdisc_code("C70793", "sponsor"), 'identifier': "333333333", 'identifierScheme': "DUNS", 'legalAddress': None}
  ]
  for item in items:
    org = factory.item(Organization, item)
    globals.cross_references.add(item['name'], org)


 


from typing import Literal, Union
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .code import Code

class Timing(ApiBaseModelWithIdNameLabelAndDesc):
  type: Code
  value: str
  valueLabel: str
  relativeToFrom: Code
  relativeFromScheduledInstanceId: str
  relativeToScheduledInstanceId: Union[str, None] = None
  windowLower: Union[str, None] = None
  windowUpper: Union[str, None] = None
  windowLabel: Union[str, None] = None
  instanceType: Literal['Timing']




from usdm_model.narrative_content import NarrativeContent
from usdm_db.fhir.from_fhir import FromFHIR
from usdm_db import USDMDb

def test_create(mocker, globals, minimal, factory):
  fhir = FromFHIR(globals.errors_and_logging)
  assert fhir is not None

def test_from_fhir_1(mocker, globals, minimal, factory):
  with open(f"tests/integration_test_files/full_1_fhir.json", 'r') as f:
    data = f.read()
  fhir = FromFHIR(globals.errors_and_logging)
  wrapper = fhir.from_fhir(data)
  assert wrapper is not None

def test_from_fhir_1(mocker, globals, minimal, factory):
  with open(f"tests/integration_test_files/full_1_fhir.json", 'r') as f:
    data = f.read()
  usdm = USDMDb()
  fhir = usdm.from_fhir(data)
  html = usdm.to_html('document')
  print(html)



import pytest
from usdm_excel.cdisc_bc_library import CDISCBCLibrary

xfail = pytest.mark.xfail

# def test_code(mocker, globals):
#     mock_id = mocker.patch("usdm_excel.id_manager.IdManager.build_id")
#     mock_id.side_effect=['Code_1']
#     item = NCIt()
#     code = item.code(code="CODE", decode="DECODE")
#     assert code.codeId == "Code_1"
#     assert code.code == "CODE"
#     assert code.codeSystem == "NCI Thesaurus"
#     assert code.codeSystemVersion == ""
#     assert code.decode == "DECODE"

@xfail
def test_create():
  assert 0

@xfail
def test_exists():
  assert 0

@xfail
def test_catalogue():
  assert 0

@xfail
def test_usdm():
  assert 0

@xfail
def test__get_package_metadata():
  assert 0

@xfail
def test__get_package_items(self):
  assert 0
  
def test__url(globals):
  item = CDISCBCLibrary(globals.errors_and_logging, globals.cdisc_ct_library, globals.id_manager)
  assert (item._url('something')) == f"{CDISCBCLibrary.API_ROOT}something"

@xfail
def test__bc_as_usdm(mocker, globals):
  assert 0

@xfail
def test__bc_property_as_usdm(self, property, codes):
  assert 0



import traceback
import pandas as pd
from usdm_excel.base_sheet import BaseSheet
from usdm_model.procedure import Procedure
from usdm_excel.globals import Globals

class StudyDesignProcedureSheet(BaseSheet):

  SHEET_NAME = 'studyDesignProcedures'
  
  def __init__(self, file_path: str, globals: Globals):
    try:
      self.procedures = []
      super().__init__(file_path=file_path, globals=globals, sheet_name=self.SHEET_NAME)
      for index, row in self.sheet.iterrows():
        xref = self.read_cell_by_name(index, "xref", default="", must_be_present=False)
        name = self.read_cell_by_name(index, ["procedureName", 'name'])
        description = self.read_cell_by_name(index, ['procedureDescription', 'description'])
        label = self.read_cell_by_name(index, 'label', default="", must_be_present=False)
        type = self.read_cell_by_name(index, "procedureType")
        code = self.read_other_code_cell_by_name(index, ['procedureCode', 'code'])
        try:
          item = Procedure(id=self.globals.id_manager.build_id(Procedure),
            name=name,
            description=description,
            label=label,
            procedureType=type, 
            code=code 
          )
        except Exception as e:
          self._general_exception(f"Failed to create Procedure object", e)
        else:
          self.procedures.append(item)
          cross_ref = xref if xref else name
          self.globals.cross_references.add(cross_ref, item)        
    except Exception as e:
      self._sheet_exception(e)


from tests.test_factory import Factory
from usdm_excel.quantity_type import QuantityType

def test_quantity_type(mocker, globals):
  test_data = [
    ('15 days', "15", 'days', 'C25301', '15 days'),
    ('1%', "1", '%', 'C25613', '1%'),
    ('  10    %', "10", '%', 'C25613', '10    %'),
  ]
  for index, test in enumerate(test_data):
    item = QuantityType(test[0], globals)
    assert(item.value) == test[1]
    assert(item.units) == test[2]
    assert(item.units_code.code) == test[3]
    assert(item.errors) == [] 
    assert(item.empty) == False
    assert(item.label) == test[4]

def test_quantity_type_empty(mocker, globals):
  test_data = [
    ('15 days', "15", 'days', 'C25301', False, '15 days'),
    ('', None, None, None, True, ''),
  ]
  for index, test in enumerate(test_data):
    item = QuantityType(test[0], globals, allow_empty=True)
    assert(item.value) == test[1]
    assert(item.units) == test[2]
    if test[3]:
      assert(item.units_code.code) == test[3]
    else:
      assert(item.units_code) == test[3]
    assert(item.errors) == [] 
    assert(item.empty) == test[4]
    assert(item.label) == test[5]

def test_quantity_type_no_units(mocker, globals):
  test_data = [
    (' 15 ', "15", None, None, False, '15'),
    (' 15 C ', "15", 'C', 'C42559', False, '15 C')
  ]
  for index, test in enumerate(test_data):
    #print(f"TEST {test}")
    item = QuantityType(test[0], globals, allow_missing_units=True)
    assert(item.value) == test[1]
    assert(item.units) == test[2]
    if test[3]:
      assert(item.units_code.code) == test[3]
    else:
      assert(item.units_code) == test[3]
    assert(item.errors) == [] 
    assert(item.empty) == test[4]
    assert(item.label) == test[5]

def test_range_type_error(mocker, globals):
  test_data = [
    (' Days', False, False, "Could not decode the quantity value ' Days'"),
    ('1', False, False, "Could not decode the quantity value, possible typographical errors '1'"),
    ('abc', False, False, "Could not decode the quantity value 'abc'"),
    ('10 slugs', False, False, "Unable to set the units code for the quantity '10 slugs'")
  ]
  for index, test in enumerate(test_data):
    item = QuantityType(test[0], globals, allow_missing_units=test[1], allow_empty=test[2])
    assert item.errors == [test[3]]



from usdm_excel.globals import Globals
from usdm_model.alias_code import AliasCode

class Alias():

  def __init__(self, globals: Globals):
    self._globals = globals

  def code(self, standard_code, aliases):
    return AliasCode(id=self._globals.id_manager.build_id(AliasCode), standardCode=standard_code, standardCodeAliases=aliases) if standard_code else None


from usdm_excel.base_sheet import BaseSheet
from usdm_model.narrative_content import NarrativeContent, NarrativeContentItem
from usdm_model.study_version import StudyVersion
from usdm_model.study_definition_document_version import StudyDefinitionDocumentVersion
from usdm_excel.globals import Globals
from usdm_excel.document.macros import Macros

class DocumentContentSheet(BaseSheet):

  SHEET_NAME = 'documentContent'
  DIV_OPEN_NS = '<div xmlns="http://www.w3.org/1999/xhtml">'
  DIV_OPEN = '<div>'
  DIV_CLOSE = '</div>'

  def __init__(self, file_path: str, globals: Globals):
    try:
      self.items = []
      self._map = {}
      super().__init__(file_path=file_path, globals=globals, sheet_name=self.SHEET_NAME, optional=True)
      if self.success:
        for index, row in self.sheet.iterrows():
          text = self.read_cell_by_name(index, 'text')
          name = self.read_cell_by_name(index, 'name')
          item = self.create_object(NarrativeContentItem, {'name': name, 'text': self._wrap_div(text)})
          if item:
            self.items.append(item)
            self.globals.cross_references.add(name, item)     
            self._map[item.id] = item
    except Exception as e:
      self._sheet_exception(e)

  # @todo: This code should probably be elsewhere
  def resolve(self, study_version: StudyVersion, document_version: StudyDefinitionDocumentVersion):
    macros = Macros(self, study_version, document_version)
    for nc in document_version.contents:
      if nc.contentItemId:
        nci = self._map[nc.contentItemId]
        nci.text = macros.resolve(nci.text)

  def _wrap_div(self, text):
    if text.startswith(self.DIV_OPEN_NS):
      return text
    elif text.startswith(self.DIV_OPEN):
      return text.replace(self.DIV_OPEN, self.DIV_OPEN_NS)
    else:
      return f'{self.DIV_OPEN_NS}{text}{self.DIV_CLOSE}'
  



import pytest
import pandas as pd
from usdm_excel.base_sheet import BaseSheet
from usdm_excel.study_definition_document.document_content_sheet import DocumentContentSheet
from usdm_excel.option_manager import Options, EmptyNoneOption
from tests.test_factory import Factory

xfail = pytest.mark.xfail

def test_create(mocker, globals):
  globals.option_manager.set(Options.EMPTY_NONE, EmptyNoneOption.EMPTY)

  mock_present = mocker.patch("usdm_excel.base_sheet.BaseSheet._sheet_present")
  mock_present.side_effect=[True]
  mock_id = mocker.patch("usdm_excel.id_manager.IdManager.build_id")
  mock_id.side_effect=['Content_1', 'Content_2', 'Content_3', 'Content_4', 'Content_5', 'Content_6', 'Content_7', 'Content_8']
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [
    ['ITEM1', 'Text 1'], 
    ['ITEM2', 'Text 1.1'], 
    ['ITEM3', 'Text 1.2'], 
    ['ITEM4', 'Text 1.2.1'], 
    ['ITEM5', 'Text 2'], 
    ['ITEM6', 'Text 2.1'], 
    ['ITEM7', 'Text 3'], 
  ]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['name', 'text'])
  content = DocumentContentSheet("", globals)
  assert len(content.items) == 7
  assert content.items[0].id == 'Content_1'
  assert content.items[0].name == 'ITEM1'
  assert content.items[0].text == '<div xmlns="http://www.w3.org/1999/xhtml">Text 1</div>'
  assert content.items[1].name == 'ITEM2'
  assert content.items[2].name == 'ITEM3'
  assert content.items[3].name == 'ITEM4'
  assert content.items[4].name == 'ITEM5'
  assert content.items[5].name == 'ITEM6'
  assert content.items[6].id == 'Content_7'
  assert content.items[6].name == 'ITEM7'

def test_create_standard_section(mocker, globals):
  mock_present = mocker.patch("usdm_excel.base_sheet.BaseSheet._sheet_present")
  mock_present.side_effect=[True]
  mock_id = mocker.patch("usdm_excel.id_manager.IdManager.build_id")
  mock_id.side_effect=['Content_1', 'Content_2', 'Content_3', 'Content_4', 'Content_5', 'Content_6', 'Content_7', 'Content_8']
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [
    ['ITEM1', 'Text 1'], 
    ['ITEM1', 'Text 1.1'], 
    ['ITEM1', '<usdm:section name="m11-title">'], 
    ['ITEM1', '<div><usdm:section name="m11-title"></div>'], 
    ['ITEM1', 'Text 2'], 
  ]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['name', 'text'])
  content = DocumentContentSheet("", globals)
  assert len(content.items) == 5
  assert content.items[0].text == '<div xmlns="http://www.w3.org/1999/xhtml">Text 1</div>'
  assert content.items[1].text == '<div xmlns="http://www.w3.org/1999/xhtml">Text 1.1</div>'
  assert content.items[2].text == '<div xmlns="http://www.w3.org/1999/xhtml"><usdm:section name="m11-title"></div>'
  assert content.items[3].text == '<div xmlns="http://www.w3.org/1999/xhtml"><usdm:section name="m11-title"></div>'
  assert content.items[4].text == '<div xmlns="http://www.w3.org/1999/xhtml">Text 2</div>'

def test_create_empty(mocker, globals):
  mock_present = mocker.patch("usdm_excel.base_sheet.BaseSheet._sheet_present")
  mock_present.side_effect=[True]
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = []
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['sectionNumber', 'name', 'sectionTitle', 'text'])
  content = DocumentContentSheet("", globals)
  assert len(content.items) == 0

def test_read_cell_by_name_error(mocker, globals):
  globals.template_manager.add('sponsor', 'studyDesignContent')
  call_parameters = []
  
  def my_add(sheet, row, column, message, level=10):
    call_parameters.append((sheet, row, column, message, level))
    return None

  mock_present = mocker.patch("usdm_excel.base_sheet.BaseSheet._sheet_present")
  mock_present.side_effect=[True]
  mock_error = mocker.patch("usdm_excel.errors_and_logging.errors.Errors.add", side_effect=my_add)
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [['Section 1']]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['name'])
  content = DocumentContentSheet("", globals)
  assert mock_error.call_count == 1
  mock_error.assert_has_calls([mocker.call('documentContent', 1, -1, "Error attempting to read cell 'text'. Exception: Failed to detect column(s) 'text' in sheet", 40)])
  


from usdm_excel.base_sheet import BaseSheet
from usdm_model.study_amendment import StudyAmendment
from usdm_model.study_amendment_impact import StudyAmendmentImpact
from usdm_excel.globals import Globals

class StudyAmendmentImpactSheet(BaseSheet):

  SHEET_NAME = 'amendmentImpact'

  def __init__(self, file_path: str, globals: Globals):
    try:
      self.items = []
      super().__init__(file_path=file_path, globals=globals, sheet_name=self.SHEET_NAME, optional=True)
      if self.success:
        for index, row in self.sheet.iterrows():
          amendment = self.read_cell_by_name(index, 'amendment')
          text = self.read_cell_by_name(index, 'text')
          substantial = self.read_boolean_cell_by_name(index, 'substantial')
          type = self.read_cdisc_klass_attribute_cell_by_name('StudyAmendmentImpact', 'type', index, 'type')
          item = self.create_object(StudyAmendmentImpact, {'text': text, 'isSubstantial': substantial, 'type': type})
          if item:
            self.items.append(item)
            parent = self.globals.cross_references.get(StudyAmendment, amendment)
            if parent:
              parent.impacts.append(item)
            else:
              column = self._get_column_index('amendment')
              self._error(index, column, f"Failed to find amendment '{amendment}'")
    except Exception as e:
      self._sheet_exception(e)



from typing import List, Literal
from .alias_code import AliasCode
from .api_base_model import ApiBaseModelWithIdNameAndLabel
from .response_code import ResponseCode
from .comment_annotation import CommentAnnotation

class BiomedicalConceptProperty(ApiBaseModelWithIdNameAndLabel):
  isRequired: bool
  isEnabled: bool
  datatype: str
  responseCodes: List[ResponseCode] = []
  code: AliasCode
  notes: List[CommentAnnotation] = []
  instanceType: Literal['BiomedicalConceptProperty']



# from usdm_excel.option_manager import option_manager
# from usdm_excel.ct_version_manager import ct_version_manager
# #from usdm_excel.id_manager import id_manager
# #from usdm_excel.cross_ref import cross_references
# from usdm_excel.errors.errors import error_manager

def clear():
  pass
  # self.globals.option_manager.clear()
  # ct_version_manager.clear()
  # id_manager.clear()
  # self.globals.cross_references.clear()
  # ct_version_manager.clear()
  # self.globals.errors_and_logging.errors().clear()



from typing import List, Literal
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from datetime import date
from .code import Code
from .geographic_scope import GeographicScope

class GovernanceDate(ApiBaseModelWithIdNameLabelAndDesc):
  type: Code
  dateValue: date
  geographicScopes: List[GeographicScope]
  instanceType: Literal['GovernanceDate']



from usdm_model.organization import Organization
from usdm_excel.base_sheet import BaseSheet

def get_organization(parent: BaseSheet, index: int) -> Organization:
  organization = None
  org_type = parent.read_cdisc_klass_attribute_cell_by_name('Organization', 'organizationType', index, ['organisationType', 'organizationType', 'type'])     
  org_id_scheme = parent.read_cell_by_name(index, ['organisationIdentifierScheme', 'organizationIdentifierScheme', 'identifierScheme'])
  org_identifier = parent.read_cell_by_name(index, ['organisationIdentifier', 'organizationIdentifier'])
  org_name = parent.read_cell_by_name(index, ['organisationName', 'organizationName', 'name'])
  org_label = parent.read_cell_by_name(index, 'label', default="", must_be_present=False)
  org_address = parent.read_address_cell_by_name(index, ['organisationAddress', 'organizationAddress', 'address'])
  if org_address:
    parent.globals.cross_references.add(org_address.id, org_address)   
  organization = parent.create_object(Organization, {'identifierScheme': org_id_scheme, 'identifier': org_identifier, 'name': org_name, 'label': org_label, 'type': org_type, 'legalAddress': org_address})
  if organization:
    parent.globals.cross_references.add(organization.name, organization)   
  return organization



from usdm_excel.base_sheet import BaseSheet
from usdm_model.population_definition import StudyDesignPopulation, StudyCohort
from usdm_model.range import Range
from usdm_model.characteristic import Characteristic
from usdm_excel.globals import Globals

class StudyDesignPopulationSheet(BaseSheet):

  SHEET_NAME = 'studyDesignPopulations'
  
  def __init__(self, file_path: str, globals: Globals):
    try:
      self._cohorts = []
      super().__init__(file_path=file_path, globals=globals, sheet_name=self.SHEET_NAME)
      self.population = None
      for index, row in self.sheet.iterrows():
        level = self.read_cell_by_name(index, 'level')
        name = self.read_cell_by_name(index, 'name')
        description = self.read_cell_by_name(index, 'description')
        label = self.read_cell_by_name(index, 'label')
        required_number = self.read_range_cell_by_name(index, "plannedCompletionNumber", require_units=False, allow_empty=True)
        recruit_number = self.read_range_cell_by_name(index, "plannedEnrollmentNumber", require_units=False, allow_empty=True)
        planned_age = self.read_range_cell_by_name(index, "plannedAge", require_units=True, allow_empty=True)
        healthy = self.read_boolean_cell_by_name(index, 'includesHealthySubjects', must_be_present=False)
        characteristics = self.read_cell_multiple_by_name(index, 'characteristics', must_be_present=False)
        codes = self._build_codes(row, index)
        if level.upper() == "MAIN":
          self.population = self._study_population(name, description, label, recruit_number, required_number, planned_age, healthy, codes)
        else:
          cohort = self._study_cohort(name, description, label, recruit_number, required_number, planned_age, healthy, codes, characteristics)
      if self.population: 
        self.population.cohorts = self._cohorts
      else:
        self._general_error(f"Not main study population detected")
    except Exception as e:
      self._sheet_exception(e)

  def _build_codes(self, row, index):
    code = self.read_cdisc_klass_attribute_cell_by_name('StudyDesignPopulation', "plannedSexOfParticipants", index, "plannedSexOfParticipants", allow_empty=True)
    return [code] if code else []

  def _study_population(self, name: str, description: str, label: str, recruit_number: Range, required_number: Range, planned_age: Range, healthy: bool, codes: list) -> StudyDesignPopulation:    
    params = {
      'name': name,
      'description': description,
      'label': label,
      'includesHealthySubjects': healthy,
      'plannedEnrollmentNumber': recruit_number,
      'plannedCompletionNumber': required_number,
      'plannedAge': planned_age,
      'plannedSex': codes
    }
    item = self.create_object(StudyDesignPopulation, params)
    if item:
      self.globals.cross_references.add(name, item)
    return item

  def _study_cohort(self, name: str, description: str, label: str, recruit_number: Range, required_number: Range, planned_age: Range, healthy: bool, codes: list, characteristics: list) -> StudyCohort:    
    characteristic_refs = self._resolve_characteristics(characteristics)
    params = {  
      'name': name,
      'description': description,
      'label': label,
      'includesHealthySubjects': healthy,
      'plannedEnrollmentNumber': recruit_number,
      'plannedCompletionNumber': required_number,
      'plannedAge': planned_age,
      'plannedSex': codes,
      'characteristics': characteristic_refs
    }
    item = self.create_object(StudyCohort, params)
    if item:
      self.globals.cross_references.add(name, item)
      self._cohorts.append(item)
    return item
      
  def _resolve_characteristics(self, names):
    results = []
    for name in names:
      object = self.globals.cross_references.get(Characteristic, name)
      if object:
        results.append(object)
      else:
        self._general_warning(f"Characterisitc '{name}' not found")
    return results



from typing import Literal, Union
from .api_base_model import ApiBaseModelWithId
from .code import Code

class StudyAmendmentReason(ApiBaseModelWithId):
  code: Code
  otherReason: Union[str, None] = None
  instanceType: Literal['StudyAmendmentReason']



import warnings
from bs4 import BeautifulSoup
from usdm_db.errors_and_logging.errors_and_logging import ErrorsAndLogging

def usdm_reference(item, attribute):
  return f'<usdm:ref klass="{item.__class__.__name__}" id="{item.id}" attribute="{attribute}"></usdm:ref>'

def get_soup(text: str, errors_and_logging: ErrorsAndLogging):
  try:
    with warnings.catch_warnings(record=True) as warning_list:
      result =  BeautifulSoup(text, 'html.parser')
    if warning_list:
      for item in warning_list:
        errors_and_logging.debug(f"Warning raised within Soup package, processing '{text}'\nMessage returned '{item.message}'")
    return result
  except Exception as e:
    errors_and_logging.exception(f"Parsing '{text}' with soup", e)
    return ""
  


from tests.mocks.mock_general import *
from tests.mocks.mock_sheet import *
from tests.mocks.mock_ids import *
from tests.mocks.mock_logging import *
from usdm_model.study_amendment_impact import StudyAmendmentImpact
from usdm_excel.study_amendment_sheet.study_amendment_impact_sheet import StudyAmendmentImpactSheet
from usdm_model.code import Code

def test_create(mocker, globals):
  sheet_data = {
    'amendment': ['A1', 'A2', 'A3'], 
    'text': ['Text One', 'Text Two', 'Text Three'], 
    'substantial': [True, False, True],
    'type': ['xxx', 'yyy', 'zzz']
  }
  expected_1 = Code(id='Code_1', code='X', codeSystem='SPONSOR', codeSystemVersion='', decode="Y")
  expected_2 = Code(id='Code_2', code='AAA', codeSystem='SPONSOR', codeSystemVersion='', decode="BBB")
  expected_3 = Code(id='Code_3', code='WWW', codeSystem='SPONSOR', codeSystemVersion='', decode="1234")
  mock_code = mocker.patch("usdm_excel.cdisc_ct.CDISCCT.code_for_attribute")
  mock_code.side_effect=[expected_1, expected_2, expected_3]
  object_data = [
    {'cls': StudyAmendmentImpact, 'data': {'text': sheet_data['text'][0], 'isSubstantial': sheet_data['substantial'][0], 'type': expected_1}},
    {'cls': StudyAmendmentImpact, 'data': {'text': sheet_data['text'][1], 'isSubstantial': sheet_data['substantial'][1], 'type': expected_2}},
    {'cls': StudyAmendmentImpact, 'data': {'text': sheet_data['text'][2], 'isSubstantial': sheet_data['substantial'][2], 'type': expected_3}}
  ]
  mock_create_object(mocker, object_data)
  mock_sheet_present(mocker)
  mock_sheet(mocker, globals, sheet_data)
  item = StudyAmendmentImpactSheet("", globals)
  assert len(item.items) == 3
  assert item.items[0].model_dump() == {
    'id': 'StudyAmendmentImpact_1',
    'instanceType': 'StudyAmendmentImpact',
    'isSubstantial': True,
    'notes': [],
    'text': 'Text One',
    'type': {
        'code': 'X',
        'codeSystem': 'SPONSOR',
        'codeSystemVersion': '',
        'decode': 'Y',
        'id': 'Code_1',
        'instanceType': 'Code',
    }
  }
  assert item.items[1].model_dump() == {
    'id': 'StudyAmendmentImpact_2',
    'instanceType': 'StudyAmendmentImpact',
    'isSubstantial': False,
    'notes': [],
    'text': 'Text Two',
    'type': {
        'code': 'AAA',
        'codeSystem': 'SPONSOR',
        'codeSystemVersion': '',
        'decode': 'BBB',
        'id': 'Code_2',
        'instanceType': 'Code',
    }
  }
  assert item.items[2].model_dump() == {
    'id': 'StudyAmendmentImpact_3',
    'instanceType': 'StudyAmendmentImpact',
    'isSubstantial': True,
    'notes': [],
    'text': 'Text Three',
    'type': {
        'code': 'WWW',
        'codeSystem': 'SPONSOR',
        'codeSystemVersion': '',
        'decode': '1234',
        'id': 'Code_3',
        'instanceType': 'Code',
    }
  }
  
def test_create_empty(mocker, globals):
  sheet_data = {}
  mock_sheet_present(mocker)
  mock_sheet(mocker, globals, sheet_data)
  item = StudyAmendmentImpactSheet("", globals)
  assert len(item.items) == 0

def test_read_cell_by_name_error(mocker, globals):
  sheet_data = {
    'amendment': ['A1'], 
    'substantial': [True],
    'type': ['Y'],
  }
  expected_1 = Code(id='Code_1', code='X', codeSystem='SPONSOR', codeSystemVersion='', decode="Y")
  mock_code = mocker.patch("usdm_excel.cdisc_ct.CDISCCT.code_for_attribute")
  mock_code.side_effect=[expected_1]
  mea = mock_error_add(mocker, [None, None, None, None, None, None])
  mock_sheet_present(mocker)
  mock_sheet(mocker, globals, sheet_data)
  item = StudyAmendmentImpactSheet("", globals)
  assert mock_called(mea, 2)
  mock_parameters_correct(mea, [mocker.call('amendmentImpact', 1, -1, "Error attempting to read cell 'text'. Exception: Failed to detect column(s) 'text' in sheet", 40), 
                                mocker.call('amendmentImpact', 1, 1, "Failed to find amendment 'A1'", 40)])



from usdm_excel.configuration_sheet import ConfigurationSheet
from usdm_excel.study_sheet.study_sheet import StudySheet
from usdm_excel.errors_and_logging.errors import Errors
from usdm_excel.globals import Globals
from usdm_excel.annotation_sheet.annotation_sheet import AnnotationSheet
from usdm_excel.abbreviation_sheet.abbreviation_sheet import AbbreviationSheet
from usdm_excel.study_identifiers_sheet.study_identifiers_sheet import StudyIdentifiersSheet
from usdm_excel.study_identifiers_sheet.study_references_sheet import StudyReferencesSheet
from usdm_excel.study_design_sheet.study_design_sheet import StudyDesignSheet
from usdm_excel.study_soa_v2_sheet.study_soa_v2_sheet import StudySoAV2Sheet
from usdm_excel.study_design_indication_sheet.study_design_indication_sheet import StudyDesignIndicationSheet
from usdm_excel.study_design_intervention_sheet.study_design_intervention_sheet import StudyDesignInterventionSheet
from usdm_excel.study_design_population_sheet.study_design_population_sheet import StudyDesignPopulationSheet
from usdm_excel.study_design_objective_endpoint_sheet.study_design_objective_endpoint_sheet import StudyDesignObjectiveEndpointSheet
from usdm_excel.study_design_estimands_sheet.study_design_estimands_sheet import StudyDesignEstimandsSheet
from usdm_excel.study_design_procedure_sheet.study_design_procedure_sheet import StudyDesignProcedureSheet
from usdm_excel.study_design_encounter_sheet.study_design_encounter_sheet import StudyDesignEncounterSheet
from usdm_excel.study_design_element_sheet.study_design_element_sheet import StudyDesignElementSheet
from usdm_excel.study_design_arm_sheet.study_design_arm_sheet import StudyDesignArmSheet
from usdm_excel.study_design_epoch_sheet.study_design_epoch_sheet import StudyDesignEpochSheet
from usdm_excel.study_design_activity_sheet.study_design_activity_sheet import StudyDesignActivitySheet
from usdm_excel.study_design_timing_sheet.study_design_timing_sheet import StudyDesignTimingSheet
from usdm_excel.study_definition_document.document_content_sheet import DocumentContentSheet
from usdm_excel.study_definition_document.document_template_sheet import DocumentTemplates
from usdm_excel.study_amendment_sheet.study_amendment_sheet import StudyAmendmentSheet
from usdm_excel.study_amendment_sheet.study_amendment_changes_sheet import StudyAmendmentChangesSheet
from usdm_excel.study_amendment_sheet.study_amendment_impact_sheet import StudyAmendmentImpactSheet
from usdm_excel.study_design_dictionary_sheet.study_design_dictionary_sheet import StudyDesignDictionarySheet
from usdm_excel.study_design_eligibility_criteria_sheet.study_design_eligibility_criteria_sheet import StudyDesignEligibilityCriteriaSheet
from usdm_excel.study_design_sites_sheet.study_design_sites_sheet import StudyDesignSitesSheet
from usdm_excel.study_design_conditions_sheet.study_design_conditions_sheet import StudyDesignConditionSheet
from usdm_excel.study_design_characteristics_sheet.study_design_characteristics_sheet import StudyDesignCharacteristicSheet
from usdm_excel.assigned_person_sheet.assigned_person_sheet import AssignedPersonSheet
from usdm_excel.study_role_sheet.study_role_sheet import StudyRoleSheet
from usdm_excel.option_manager import Options, EmptyNoneOption
from usdm_excel.cdisc_ct import CDISCCT
from usdm_excel.other_ct import OtherCT
from usdm_model.study import Study
from usdm_model.study_version import StudyVersion
from usdm_model.study_definition_document_version import StudyDefinitionDocumentVersion
from usdm_model.study_definition_document import StudyDefinitionDocument
from usdm_model.wrapper import Wrapper
from usdm_info import __model_version__ as usdm_version, __package_version__ as system_version

class USDMExcel():

  SYSTEM_NAME = "CDISC USDM E2J"
  STUDY_VERSION_DATE = 'study_version'
  PROTOCOL_VERSION_DATE = 'protocol_document'

  def __init__(self, file_path):
    self._globals = Globals()
    self._file_path = file_path

  def execute(self):
    self._globals.create()
    return self._process()

  def errors(self):
    return self._globals.errors_and_logging.errors().dump(Errors.WARNING)

  def is_m11(self):
    return self._globals.template_manager.includes('M11')

  def was_m11(self):
    self._globals.errors_and_logging.deprecated("Method 'was_m11' deprecated, use 'is_m11' going forward")
    return self._globals.template_manager.includes('M11')
  
  def default_template(self):
    return self._globals.template_manager.default_template

  def templates(self):
    return self._globals.template_manager.templates()

  def _process(self):
    try:
    
      # Read the configuration. Override the template if requested and present, otherwise leave as configured
      self.configuration = ConfigurationSheet(self._file_path, self._globals)

      # Process all the remaining sheets
      self.notes = AnnotationSheet(self._file_path, self._globals)
      self.abbreviations = AbbreviationSheet(self._file_path, self._globals)
      self.study = StudySheet(self._file_path, self._globals)
      self.timings = StudyDesignTimingSheet(self._file_path, self._globals)
      self.study_amendments = StudyAmendmentSheet(self._file_path, self._globals)
      self.study_amendment_impacts = StudyAmendmentImpactSheet(self._file_path, self._globals)
      self.study_amendment_changes = StudyAmendmentChangesSheet(self._file_path, self._globals)
      self.study_identifiers = StudyIdentifiersSheet(self._file_path, self._globals)
      self.study_references = StudyReferencesSheet(self._file_path, self._globals)
      self.procedures = StudyDesignProcedureSheet(self._file_path, self._globals)
      self.encounters = StudyDesignEncounterSheet(self._file_path, self._globals)
      self.elements = StudyDesignElementSheet(self._file_path, self._globals)
      self.arms = StudyDesignArmSheet(self._file_path, self._globals)
      self.epochs = StudyDesignEpochSheet(self._file_path, self._globals)
      self.activities = StudyDesignActivitySheet(self._file_path, self._globals)
      self.study_design = StudyDesignSheet(self._file_path, self._globals)
      self._process_soa()
      self.indications = StudyDesignIndicationSheet(self._file_path, self._globals)
      self.interventions = StudyDesignInterventionSheet(self._file_path, self._globals)
      self.study_characteristics = StudyDesignCharacteristicSheet(self._file_path, self._globals)
      self.study_population = StudyDesignPopulationSheet(self._file_path, self._globals)

      self.contents = DocumentContentSheet(self._file_path, self._globals)
      self.doc_templates = DocumentTemplates(self._file_path, self._globals)

      self.dictionaries = StudyDesignDictionarySheet(self._file_path, self._globals)
      self.oe = StudyDesignObjectiveEndpointSheet(self._file_path, self._globals)
      self.eligibility_criteria = StudyDesignEligibilityCriteriaSheet(self._file_path, self._globals)
      self.estimands = StudyDesignEstimandsSheet(self._file_path, self._globals)
      self.sites = StudyDesignSitesSheet(self._file_path, self._globals)
      self.conditions = StudyDesignConditionSheet(self._file_path, self._globals)

      self.people = AssignedPersonSheet(self._file_path, self._globals)
      self.roles = StudyRoleSheet(self._file_path, self._globals)

      # Study Design assembly
      study_design = self.study_design.study_designs[0]
      study_design.scheduleTimelines.append(self.soa.timeline)
      study_design.encounters = self.encounters.items
      study_design.activities = self.soa.activities
      activity_ids = [item.id for item in study_design.activities]
      study_design.biomedicalConcepts = self.soa.biomedical_concepts
      study_design.bcSurrogates = self.soa.biomedical_concept_surrogates
      for key, tl in self.study.timelines.items():
        study_design.scheduleTimelines.append(tl.timeline)
        for activity in tl.activities:
          if activity.id not in activity_ids:
            study_design.activities.append(activity)
            activity_ids.append(activity.id)
        study_design.biomedicalConcepts += tl.biomedical_concepts
        study_design.bcSurrogates += tl.biomedical_concept_surrogates
      self._double_link(study_design.activities, 'previousId', 'nextId')
      study_design.indications = self.indications.items
      study_design.studyInterventions = self.interventions.items
      study_design.population = self.study_population.population
      study_design.objectives = self.oe.objectives
      study_design.estimands = self.estimands.estimands
      study_design.population.criterionIds = [x.id for x in self.eligibility_criteria.items]
      study_design.dictionaries = self.dictionaries.items
      study_design.conditions = self.conditions.items

      self.definition_documents = []
      self.definition_document_version_ids = []
      self.definition_document_version = []
      for template in self.doc_templates.items:
        # Final assembly
        try:
          definition_document_version = StudyDefinitionDocumentVersion(
            id=self._globals.id_manager.build_id(StudyDefinitionDocumentVersion), 
            version=self.study.protocol_version,
            status=self.study.protocol_status,
            dateValues=self.study.dates[self.PROTOCOL_VERSION_DATE],
            contents=template.items
            )
          self.definition_document_version.append(definition_document_version)
          self.definition_document_version_ids.append(definition_document_version.id)
          self._globals.cross_references.add(definition_document_version.id, definition_document_version)
        except Exception as e:
         self._globals.errors_and_logging.exception(f"Error creating StudyDefinitionDocumentVersion object", e)

        try:
          definition_document = StudyDefinitionDocument(
            id=self._globals.id_manager.build_id(StudyDefinitionDocument), 
            name=f"Protocol_Document_{self.study.name}", 
            versions=[definition_document_version],
            language=OtherCT(self._globals).code("en", "ISO", '1', "English"),
            type=CDISCCT(self._globals).code("C12345", "Decode"),
            templateName=template.name
          )
          self.definition_documents.append(definition_document)
        except Exception as e:
          self._globals.errors_and_logging.exception(f"Error creating StudyDefinitionDocument object", e)

      try:
        self.study_version = StudyVersion(
          id=self._globals.id_manager.build_id(StudyVersion),
          versionIdentifier=self.study.version,
          studyType=self.study.type,
          studyPhase=self.study.phase,
          businessTherapeuticAreas=self.study.therapeutic_areas,
          rationale=self.study.rationale,
          studyIdentifiers=self.study_identifiers.identifiers,
          referenceIdentifiers=self.study_references.items,
          documentVersionIds=self.definition_document_version_ids,
          studyDesigns=self.study_design.study_designs,
          dateValues=self.study.dates[self.STUDY_VERSION_DATE],
          amendments=self.study_amendments.items,
          titles=self.study.titles,
          criteria=self.eligibility_criteria.items,
          narrativeContentItems=self.contents.items,
          abbreviations=self.abbreviations.items,
          organizations=self.study_references.organizations + self.study_identifiers.organizations + self.sites.organizations,
          roles=self.roles.items
        )
        self._globals.cross_references.add(self.study_version.id, self.study_version)
      except Exception as e:
        self._globals.errors_and_logging.exception(f"Error creating StudyVersion object", e)

      try:
        self.study = Study(
          id=None, # No Id, will be allocated a UUID
          name=f"Study_{self.study.name}", 
          versions=[self.study_version],
          documentedBy=self.definition_documents
        )
        self._globals.cross_references.add("STUDY", self.study)
        for document in self.definition_document_version:
          self.contents.resolve(self.study_version, document) # Now we have full study, resolve references in the content
      except Exception as e:
        self._globals.errors_and_logging.exception(f"Error creating Study object", e)

      return Wrapper(study=self.study, usdmVersion=usdm_version, systemName=self.SYSTEM_NAME, systemVersion=system_version)
 
    except Exception as e:
      self._globals.errors_and_logging.exception(f"Error processing Excel workbook", e)
      return None

  def _process_soa(self):
    tls = []
    for timeline in self.study_design.other_timelines:
      tl = StudySoAV2Sheet(self._file_path, self._globals, timeline, False)
      tls.append(tl)
      self.study.timelines[timeline] = tl
      self._globals.cross_references.add(timeline, tl.timeline)
    self.soa = StudySoAV2Sheet(self._file_path, self._globals, self.study_design.main_timeline, True)
    #print(f"XREF: {self.study_design.main_timeline}")
    self._globals.cross_references.add(self.study_design.main_timeline, self.soa.timeline)
    tls.append(self.soa)
    self._set_timing_references(tls)
    self._check_timing_references(tls)

  def _check_timing_references(self, tls):
    timing_check = {}
    for timing in self.timings.items:
      timing_check[timing.name] = None
    for tl in tls:
      tl_items = tl.check_timing_references(self.timings.items, timing_check)
      tl.timeline.timings = tl_items
    for timing in self.timings.items:
      if not timing_check[timing.name]:
        self._globals.errors_and_logging.error(f"Timing with name '{timing.name}' not referenced")

  def _set_timing_references(self, tls):
    for timing in self.timings.items:
      found = {'from': False, 'to': False}
      for tl in tls:
        if not found['from']:
          instance = tl.timing_match(timing.relativeFromScheduledInstanceId)
          if instance:
            item = instance.item
            timing.relativeFromScheduledInstanceId = item.id
            found['from'] = True
        if not found['to']:
          instance = tl.timing_match(timing.relativeToScheduledInstanceId)
          if instance:
            item = instance.item
            timing.relativeToScheduledInstanceId = item.id
            found['to'] = True
      if not found['from']:
        self._globals.errors_and_logging.error(f"Unable to find timing 'from' reference with name {timing.relativeFromScheduledInstanceId}")
      if not found['to']:
        self._globals.errors_and_logging.error(f"Unable to find timing 'to' reference with name {timing.relativeToScheduledInstanceId}")

  def _double_link(self, items, prev, next):
    try: 
      for idx, item in enumerate(items):
        if idx == 0:
          if self._globals.option_manager.get(Options.EMPTY_NONE) == EmptyNoneOption.EMPTY.value:
            setattr(item, prev, "")
          else:
            setattr(item, prev, None)
        else:
          the_id = getattr(items[idx-1], 'id')
          setattr(item, prev, the_id)
        if idx == len(items)-1:  
          if self._globals.option_manager.get(Options.EMPTY_NONE) == EmptyNoneOption.EMPTY.value:
            setattr(item, next, "")
          else:
            setattr(item, next, None)
        else:
          the_id = getattr(items[idx+1], 'id')
          setattr(item, next, the_id)
    except Exception as e:
      self._globals.errors_and_logging.exception(f"Error in double_link: {items}", e)



from usdm_excel.globals import Globals
from usdm_excel.syntax_template_sheet import SyntaxTemplateSheet
from usdm_model.characteristic import Characteristic

class StudyDesignCharacteristicSheet(SyntaxTemplateSheet):

  SHEET_NAME = 'studyDesignCharacteristics'
  
  def __init__(self, file_path: str, globals: Globals):
    try:
      self.items = []
      super().__init__(file_path=file_path, globals=globals, sheet_name=self.SHEET_NAME, optional=True)
      if self.success:
        for index, row in self.sheet.iterrows():
          name = self.read_cell_by_name(index, 'name')
          description = self.read_cell_by_name(index, 'description')
          label = self.read_cell_by_name(index, 'label')
          text = self.read_cell_by_name(index, 'text')
          dictionary_name = self.read_cell_by_name(index, 'dictionary')
          self._validate_references(index, 'text', text, dictionary_name)
          item = self._characteristic(name, description, label, text, dictionary_name)
          if item:
            self.globals.cross_references.add(item.name, item)
            self.items.append(item)
    except Exception as e:
      self._sheet_exception(e)

  def _characteristic(self, name: str, description: str, label: str, text: str, dictionary_name: str) -> Characteristic:
    dictionary_id = self._get_dictionary_id(dictionary_name)
    params = {'name': name, 'description': description, 'label': label, 'text': text, 'dictionaryId': dictionary_id}
    return self.create_object(Characteristic, params)



class Conditons():

  def __init__(self, conditon_info):
    self.items = []
    self.errors = []
    if conditon_info:
      parts = conditon_info.split(",")
      for part in parts:
        name_value = part.split(":")
        if len(name_value) == 2:
          name = self._remove_unprintable(name_value[0])
          condition = self._remove_unprintable(name_value[1])
          self.items.append({'name': name, 'condition': condition})
        else:
          self.errors.append(f"Could not decode a condition, no ':' found in '{part}'")
    else:
      pass # Empty, this is OK

  def _remove_unprintable(self, text):
    return ''.join(c for c in text if c.isprintable()).strip()


import re
import traceback
from usdm_excel.base_sheet import BaseSheet
from usdm_excel.iso_8601_duration import ISO8601Duration
from usdm_excel.study_design_timing_sheet.window_type import WindowType
from usdm_excel.cdisc_ct import CDISCCT
from usdm_model.timing import Timing
from usdm_excel.globals import Globals

class StudyDesignTimingSheet(BaseSheet):

  SHEET_NAME = 'studyDesignTiming'

  def __init__(self, file_path: str, globals: Globals):
    try:
      self.items = []
      super().__init__(file_path=file_path, globals=globals, sheet_name=self.SHEET_NAME, optional=True)
      if self.success:
        for index, row in self.sheet.iterrows():
          name = self.read_cell_by_name(index, 'name')
          description = self.read_cell_by_name(index, 'description')
          label = self.read_cell_by_name(index, 'label')
          type = self._set_type(self.read_cell_by_name(index, 'type'))
          from_name = self.read_cell_by_name(index, 'from')
          to_name = self.read_cell_by_name(index, 'to')
          timing_label = self.read_cell_by_name(index, 'timingValue')
          timing_value = self._set_text_and_encoded(self.read_cell_by_name(index, 'timingValue'))
          to_from_type = self._set_to_from_type(self.read_cell_by_name(index, 'toFrom'))
          window = WindowType(self.read_cell_by_name(index, 'window'), self.globals)
          if window.errors:
            self._add_errors(window.errors, index, self._get_column_index('window'))
          try:
            item = Timing(
              id=self.globals.id_manager.build_id(Timing),
              type=type,
              value=timing_value,
              valueLabel=timing_label,
              name=name,
              description=description,
              label=label,
              relativeToFrom=to_from_type,
              windowLabel=window.label,
              windowLower=window.lower,
              windowUpper=window.upper,
              relativeFromScheduledInstanceId=from_name,
              relativeToScheduledInstanceId=to_name
            )
          except Exception as e:
            self._general_error(f"Failed to create Timing object", e)
          else:
            self.globals.cross_references.add(name, item)
            self.items.append(item)
    except Exception as e:
      self._sheet_exception(e)

  def _set_text_and_encoded(self, duration):
    the_duration = duration.strip()
    original_duration = the_duration
    for char in ['+', '-']:
      if char in the_duration:
        the_duration = the_duration.replace(char, "")
        self._log_warning(f"Ignoring '{char}' in {original_duration}")
    duration_parts = re.findall(r"[^\W\d_]+|\d+", the_duration)
    if len(duration_parts) == 2:
      try:
        return ISO8601Duration().encode(duration_parts[0].strip(), duration_parts[1].strip())
      except Exception as e:
        self._general_error(f"Could not decode the duration value '{the_duration}'")
    else:
      self._general_error(f"Could not decode the duration value, no value and units found in '{the_duration}'")

  def _set_type(self, text):
    type_code = {
      "FIXED": {'c_code': 'C201358', 'pt': 'Fixed Reference'},
      "AFTER": {'c_code': 'C201356', 'pt': 'After'},
      "BEFORE": {'c_code': 'C201357', 'pt': 'Before'}
    }   
    key = text.strip().upper()
    return CDISCCT(self.globals).code(type_code[key]['c_code'], type_code[key]['pt'])

  def _set_to_from_type(self, text):
    type_code = {
      "S2S": {'c_code': 'C201355', 'pt': 'Start to Start'},
      "S2E": {'c_code': 'C201354', 'pt': 'Start to End'},
      "E2S": {'c_code': 'C201353', 'pt': 'End to Start'},
      "E2E": {'c_code': 'C201352', 'pt': 'End to End'},
    }    
    key = "S2S" if not text else text.strip().upper()
    return CDISCCT(self.globals).code(type_code[key]['c_code'], type_code[key]['pt'])



from usdm_excel.base_sheet import BaseSheet
from usdm_model.study_amendment import StudyAmendment
from usdm_model.study_change import StudyChange
from usdm_model.document_content_reference import DocumentContentReference
from usdm_excel.globals import Globals

class StudyAmendmentChangesSheet(BaseSheet):

  SHEET_NAME = 'amendmentChanges'

  def __init__(self, file_path: str, globals: Globals):
    try:
      self.items = []
      super().__init__(file_path=file_path, globals=globals, sheet_name=self.SHEET_NAME, optional=True)
      if self.success:
        for index, row in self.sheet.iterrows():
          amendment = self.read_cell_by_name(index, 'amendment')
          name = self.read_cell_by_name(index, ['name'])
          description = self.read_cell_by_name(index, ['description'])
          label = self.read_cell_by_name(index, 'label', default="", must_be_present=False)
          rationale = self.read_cell_by_name(index, 'rationale')
          summary = self.read_cell_by_name(index, 'summary')
          sections = self._section_list(index)
          item = self.create_object(StudyChange, {'name': name, 'description': description, 'label': label, 'rationale': rationale, 'summary': summary, 'changedSections': sections})
          if item:
            self.items.append(item)
            parent = self.globals.cross_references.get(StudyAmendment, amendment)
            if parent:
              parent.changes.append(item)
            else:
              self._error(index, self._get_column_index('amendment'), f"Failed to find amendment '{amendment}'")
                
    except Exception as e:
      self._sheet_exception(e)

  def _section_list(self, index):
    result = []
    section_list = self.read_cell_multiple_by_name(index, 'sections')
    for section in section_list:
      parts = section.split(':')
      if len(parts) == 2:
        ref = self.create_object(DocumentContentReference, {'sectionNumber': parts[0].strip(), 'sectionTitle': parts[1].strip()})
        if ref:
          result.append(ref)
      else:
        self._error(index, self._get_column_index('sections'), f"Could not decode section reference '{section}'.")
    return result


import base64
from tests.test_data_factory import MinimalStudy
from usdm_model.eligibility_criterion import EligibilityCriterion
from usdm_model.biomedical_concept import BiomedicalConcept
from usdm_model.activity import Activity
from usdm_model.abbreviation import Abbreviation
from usdm_excel.document.macros import Macros
from usdm_excel.globals import Globals
from tests.test_factory import Factory

def create_criteria(factory: Factory, minimal: MinimalStudy):
  INCLUSION = factory.cdisc_code('C25532', 'Inc')
  EXCLUSION = factory.cdisc_code('C25370', 'Exc')
  item_list = [
    {'name': 'IE1', 'label': '', 'description': '', 'text': 'Only perform at baseline', 
     'dictionaryId': None, 'category': INCLUSION, 'identifier': '01', 'nextId': None, 'previousId': None, 'contextId': None
    },
    {'name': 'IE2', 'label': '', 'description': '', 'text': '<p>Only perform on males</p>', 
    'dictionaryId': None, 'category': INCLUSION, 'identifier': '02', 'nextId': None, 'previousId': None, 'contextId': None
    },
  ]
  results = factory.set(EligibilityCriterion, item_list)
  #print(f"RESULTS: {results}")
  for criterion in results:
    minimal.population.criterionIds.append(criterion.id)
  minimal.study.versions[0].criteria = results
  return results

def create_bc(factory: Factory, globals: Globals):
  code = factory.cdisc_dummy()
  alias_code = factory.alias_code(code)
  bc = factory.item(BiomedicalConcept, {'name': 'height', 'label': 'bc name', 'reference': 'something', 'code': alias_code, 'synonyms': ['body height', 'Standing height']})
  activity = factory.item(Activity, {'name': 'vitals' ,'biomedicalConceptIds': [bc.id]})
  globals.cross_references.add(bc.id, bc)
  globals.cross_references.add(activity.name, activity)

def create_abbreviations(factory: Factory, globals: Globals):
  abbreviations = [
    {'abbreviatedText': 'XXX', 'expandedText': 'X is X very X long'},
    {'abbreviatedText': 'ECG', 'expandedText': 'Electrocardiogram'},
    {'abbreviatedText': 'AD', 'expandedText': "Alzheimer's"}
  ]
  for abbreviation in abbreviations:
    item = factory.item(Abbreviation, abbreviation)
    globals.cross_references.add(item.abbreviatedText, item)

def get_instance(mocker, globals: Globals, factory: Factory, minimal: MinimalStudy):
  globals.id_manager.clear()
  criteria = create_criteria(factory, minimal)
  bs = factory.base_sheet(mocker)
  macro = Macros(bs, minimal.study_version, minimal.study_definition_document_version)
  return macro

def test_create(mocker, globals, factory, minimal):
  macro = get_instance(mocker, globals, factory, minimal)
  assert macro is not None

def test_resolve_ok(mocker, globals, factory, minimal):
  macro = get_instance(mocker, globals, factory, minimal)
  result = macro.resolve('<usdm:macro id="section" name="inclusion"/>')
  expected = '<table class="table"><tr><td>01</td><td><usdm:ref attribute="text" '\
    'id="EligibilityCriterion_1" klass="EligibilityCriterion"></usdm:ref></td></tr><tr><td>02'\
    '</td><td><usdm:ref attribute="text" id="EligibilityCriterion_2" klass="EligibilityCriterion">'\
    '</usdm:ref></td></tr></table>'
  assert result == expected

def test_invalid_method(mocker, globals, factory, minimal):
  macro = get_instance(mocker, globals, factory, minimal)
  result = macro.resolve('<usdm:macro id="xxx" name="inclusion"/>')
  expected = 'Missing content: invalid method name'
  assert result == expected

def test_invalid_exception(mocker, globals, factory, minimal):
  macro = get_instance(mocker, globals, factory, minimal)
  result = macro.resolve('<usdm:macro name="inclusion"/>')
  expected = 'Missing content: exception'
  assert result == expected

def test_bc_ok_name(mocker, globals, factory, minimal):
  macro = get_instance(mocker, globals, factory, minimal)
  create_bc(factory, globals)
  result = macro.resolve('<usdm:macro id="bc" name="height" activity="vitals"/>')
  expected = '<usdm:ref attribute="label" id="BiomedicalConcept_1" klass="BiomedicalConcept"></usdm:ref>'
  assert result == expected

def test_bc_ok_synonym_1(mocker, globals, factory, minimal):
  macro = get_instance(mocker, globals, factory, minimal)
  create_bc(factory, globals)
  result = macro.resolve('<usdm:macro id="bc" name="body HEIGHT" activity="vitals"/>')
  expected = '<usdm:ref attribute="label" id="BiomedicalConcept_1" klass="BiomedicalConcept"></usdm:ref>'
  assert result == expected

def test_bc_ok_synonym_2(mocker, globals, factory, minimal):
  macro = get_instance(mocker, globals, factory, minimal)
  create_bc(factory, globals)
  result = macro.resolve('<usdm:macro id="bc" name=" standing height " activity="vitals"/>')
  expected = '<usdm:ref attribute="label" id="BiomedicalConcept_1" klass="BiomedicalConcept"></usdm:ref>'
  assert result == expected

def test_bc_error(mocker, globals, factory, minimal):
  macro = get_instance(mocker, globals, factory, minimal)
  create_bc(factory, globals)
  result = macro.resolve('<usdm:macro id="bc" name="heightX" activity="vitals"/>')
  expected = 'Missing BC: failed to find BC in activity'
  assert result == expected

def test_activity_error(mocker, globals, factory, minimal):
  macro = get_instance(mocker, globals, factory, minimal)
  create_bc(factory, globals)
  result = macro.resolve('<usdm:macro id="bc" name="height" activity="vitalsX"/>')
  expected = "Missing activity: failed to find activity 'vitalsX'"
  assert result == expected

def test_image_ok(mocker, globals, factory, minimal):
  mock_image = mocker.patch("usdm_excel.document.macros.Macros._encode_image")
  mock_image.side_effect=[base64.b64encode(str.encode('01234567890ABCDEF'))]
  macro = get_instance(mocker, globals, factory, minimal)
  result = macro.resolve('<usdm:macro id="image" file="xxx.png" type="png"/>')
  expected = '<img alt="Alt text" src="data:image/png;base64,MDEyMzQ1Njc4OTBBQkNERUY="/>'
  assert result == expected

def test_image_error(mocker, globals, factory, minimal):
  macro = get_instance(mocker, globals, factory, minimal)
  result = macro.resolve('<usdm:macro id="image" file="xxx.png" type="png"/>')
  expected = "Missing image: failed to insert image 'xxx.png'"
  assert result == expected

def test_abbreviation_ok(mocker, globals, factory, minimal):
  macro = get_instance(mocker, globals, factory, minimal)
  create_abbreviations(factory, globals)
  result = macro.resolve('<usdm:macro id="abbreviations" items="XXX"/>')
  expected = '<usdm:ref attribute="abbreviatedText" id="Abbreviation_1" klass="Abbreviation"></usdm:ref> = <usdm:ref attribute="expandedText" id="Abbreviation_1" klass="Abbreviation"></usdm:ref>'
  assert result == expected

def test_abbreviation_separator(mocker, globals, factory, minimal):
  macro = get_instance(mocker, globals, factory, minimal)
  create_abbreviations(factory, globals)
  result = macro.resolve('<usdm:macro id="abbreviations" items="XXX, ECG" separator=";"/>')
  expected = '<usdm:ref attribute="abbreviatedText" id="Abbreviation_1" klass="Abbreviation"></usdm:ref> = <usdm:ref attribute="expandedText" id="Abbreviation_1" klass="Abbreviation"></usdm:ref>; <usdm:ref attribute="abbreviatedText" id="Abbreviation_2" klass="Abbreviation"></usdm:ref> = <usdm:ref attribute="expandedText" id="Abbreviation_2" klass="Abbreviation"></usdm:ref>'
  assert result == expected

def test_abbreviation_multiple_ok(mocker, globals, factory, minimal):
  macro = get_instance(mocker, globals, factory, minimal)
  create_abbreviations(factory, globals)
  result = macro.resolve('<usdm:macro id="abbreviations" items="XXX,  ECG,AD"/>')
  expected = '<usdm:ref attribute="abbreviatedText" id="Abbreviation_1" klass="Abbreviation"></usdm:ref> = <usdm:ref attribute="expandedText" id="Abbreviation_1" klass="Abbreviation"></usdm:ref>, <usdm:ref attribute="abbreviatedText" id="Abbreviation_2" klass="Abbreviation"></usdm:ref> = <usdm:ref attribute="expandedText" id="Abbreviation_2" klass="Abbreviation"></usdm:ref>, <usdm:ref attribute="abbreviatedText" id="Abbreviation_3" klass="Abbreviation"></usdm:ref> = <usdm:ref attribute="expandedText" id="Abbreviation_3" klass="Abbreviation"></usdm:ref>'
  assert result == expected

def test_abbreviation_error(mocker, globals, factory, minimal):
  macro = get_instance(mocker, globals, factory, minimal)
  result = macro.resolve('<usdm:macro id="abbreviations" items="AE"/>')
  expected = "Missing abbreviation: failed to find abbreviation 'AE'"
  assert result == expected




import logging
from yattag import Doc
from usdm_db.cross_reference import CrossReference
from usdm_db.errors_and_logging.errors_and_logging import ErrorsAndLogging
from usdm_model.scheduled_instance import ScheduledActivityInstance, ScheduledDecisionInstance, ScheduledInstance
from usdm_model.schedule_timeline_exit import ScheduleTimelineExit

class Timeline():

  FULL = "full"
  BODY = "body"

  def __init__(self, study):
    self._errors_and_logging = ErrorsAndLogging()
    self._study = study
    self._cross_ref = CrossReference(study, self._errors_and_logging)

  def to_html(self, level=FULL):
    try:
      doc = Doc()
      study_design = self._study.versions[0].studyDesigns[0]
      if level == self.BODY:
        self._body(doc, study_design)
      else:
        self._full(doc, study_design)
      return doc.getvalue()
    except Exception as e:
      self._errors_and_logging.exception(f"Failed generating HTML page at level '{level}'", e)

  def _full(self, doc, study_design):
    doc.asis('<!DOCTYPE html>')
    with doc.tag('html'):
      with doc.tag('head'):
        pass
      with doc.tag('body'):
        self._body(doc, study_design)
  
  def _body(self, doc, study_design):
    for timeline in study_design.scheduleTimelines:
      timings = timeline.timings
      with doc.tag(f'h1'):
        doc.asis(f'{timeline.name}')
      with doc.tag('pre', klass='mermaid'):
        doc.asis('\ngraph LR\n')
        doc.asis(f'{timeline.id}([{timeline.entryCondition}])\n')
        instance = self._cross_ref.get(ScheduledActivityInstance, timeline.entryId)
        if instance.instanceType == ScheduledActivityInstance.__name__: 
          doc.asis(f'{instance.id}(ScheduledActivityInstance)\n')
        else:
          doc.asis(f'{instance.id}{{{{ScheduledDecisionInstance}}}}\n')
        doc.asis(f'{timeline.id} -->|first| {instance.id}\n')
        prev_instance = instance
        instance = self._cross_ref.get(ScheduledActivityInstance, instance.defaultConditionId)
        while instance:
          if instance.instanceType == ScheduledActivityInstance.__name__: 
            doc.asis(f'{instance.id}(ScheduledActivityInstance)\n')
          else:
            doc.asis(f'{instance.id}{{{{ScheduledDecisionInstance}}}}\n')
            for condition in instance.conditionAssignments:
              doc.asis(f'{instance.id} -->|{condition.condition}| {condition.conditionTargetId}\n') 
          doc.asis(f'{prev_instance.id} -->|default| {instance.id}\n')      
          prev_instance = instance
          instance = self._get_cross_reference(prev_instance.defaultConditionId)
        exit = self._cross_ref.get(ScheduleTimelineExit, prev_instance.timelineExitId)
        doc.asis(f'{exit.id}([Exit])\n')
        doc.asis(f'{prev_instance.id} -->|exit| {exit.id}\n')      
        for timing in timings:
          doc.asis(f'{timing.id}(({timing.label}\n{timing.type.decode}\n{timing.value}\n{timing.windowLower}..{timing.windowUpper}))\n')            
          doc.asis(f'{timing.id} -->|from| {timing.relativeFromScheduledInstanceId}\n')      
          doc.asis(f'{timing.id} -->|to| {timing.relativeToScheduledInstanceId}\n')      
    with doc.tag('script', type='module'):
      doc.asis("import mermaid from 'https://cdn.jsdelivr.net/npm/mermaid@10/dist/mermaid.esm.min.mjs';\n")   
      doc.asis("mermaid.initialize({ startOnLoad: true });\n")   

  def _get_cross_reference(self, id):
    for klass in [ScheduledActivityInstance, ScheduledDecisionInstance]:
      instance = self._cross_ref.get(klass, id)
      if instance:
        return instance 
    return None





from yattag import Doc
from usdm_model.eligibility_criterion import EligibilityCriterion
from usdm_model.narrative_content import NarrativeContent, NarrativeContentItem
from usdm_db.document.document import Document
from tests.test_factory import Factory
from bs4 import BeautifulSoup

def create_criteria(factory, minimal):
  INCLUSION = factory.cdisc_code('C25532', 'Inc')
  EXCLUSION = factory.cdisc_code('C25370', 'Exc')
  item_list = [
    {'name': 'IE1', 'label': '', 'description': '', 'text': 'Only perform at baseline', 
     'dictionaryId': None, 'category': INCLUSION, 'identifier': '01', 'nextId': None, 'previousId': None, 'contextId': None
    },
    {'name': 'IE2', 'label': '', 'description': '', 'text': '<p>Only perform on males</p>', 
    'dictionaryId': None, 'category': INCLUSION, 'identifier': '02', 'nextId': None, 'previousId': None, 'contextId': None
    },
  ]
  results = factory.set(EligibilityCriterion, item_list)
  #print(f"RESULTS: {results}")
  for criterion in results:
    minimal.population.criterionIds.append(criterion.id)
  minimal.study.versions[0].criteria = results
  return results

def test_wrap_in_span_and_modal_1(mocker, globals, minimal, factory):
  criteria = create_criteria(factory, minimal)
  document = Document(minimal.study, "sponsor", globals.errors_and_logging)
  soup =  BeautifulSoup('Before<usdm:ref id="EligibilityCriterion_1" klass="EligibilityCriterion" attribute="text"/>After', 'html.parser')
  ref = soup(['usdm:ref'])[0]
  document._wrap_in_span_and_modal(soup, ref, "1234")
  assert str(soup) == 'Before<span class="usdm-highlight">\n<a class="link-dark usdm-highlight-link" data-bs-target="#usdmContent1" data-bs-toggle="modal" style="font-size: 12px;">\n<i class="ps-1 pe-1 bi bi-info-circle"></i>\n</a>\n1234\n<div class="modal fade" id="usdmContent1" tabindex="-1">\n<div class="modal-dialog">\n<div class="modal-content">\n<div class="modal-header">\n<h5 class="modal-title">Included using \'usdm:ref\'</h5>\n<button aria-label="Close" class="btn-close" data-bs-dismiss="modal" type="button"></button>\n</div>\n<div class="modal-body">\n              Attributes: <b>\'id\':</b> \'EligibilityCriterion_1\', <b>\'klass\':</b> \'EligibilityCriterion\', <b>\'attribute\':</b> \'text\'\n            </div>\n</div>\n</div>\n</div>\n</span>After'

def test_replace_and_highlight_1(mocker, globals, minimal, factory):
  criteria = create_criteria(factory, minimal)
  document = Document(minimal.study, "sponsor", globals.errors_and_logging)
  soup =  BeautifulSoup('Before<usdm:ref id="EligibilityCriterion_1" klass="EligibilityCriterion" attribute="text"/>After', 'html.parser')
  ref = soup(['usdm:ref'])[0]
  result = document._replace_and_highlight(soup, ref, ' ccc ', False)
  assert str(soup) == 'Before ccc After'

def test_replace_and_highlight_2(mocker, globals, minimal, factory):
  criteria = create_criteria(factory, minimal)
  document = Document(minimal.study, "sponsor", globals.errors_and_logging)
  soup =  BeautifulSoup('Before<usdm:ref id="EligibilityCriterion_1" klass="EligibilityCriterion" attribute="text"/>After', 'html.parser')
  ref = soup(['usdm:ref'])[0]
  result = document._replace_and_highlight(soup, ref, ' TEXT HERE ', True)
  assert str(soup) == 'Before<span class="usdm-highlight">\n<a class="link-dark usdm-highlight-link" data-bs-target="#usdmContent1" data-bs-toggle="modal" style="font-size: 12px;">\n<i class="ps-1 pe-1 bi bi-info-circle"></i>\n</a>\n TEXT HERE \n<div class="modal fade" id="usdmContent1" tabindex="-1">\n<div class="modal-dialog">\n<div class="modal-content">\n<div class="modal-header">\n<h5 class="modal-title">Included using \'usdm:ref\'</h5>\n<button aria-label="Close" class="btn-close" data-bs-dismiss="modal" type="button"></button>\n</div>\n<div class="modal-body">\n              Attributes: <b>\'id\':</b> \'EligibilityCriterion_1\', <b>\'klass\':</b> \'EligibilityCriterion\', <b>\'attribute\':</b> \'text\'\n            </div>\n</div>\n</div>\n</div>\n</span>After'

def test_translate_rererences_1(mocker, globals, minimal, factory):
  criteria = create_criteria(factory, minimal)
  document = Document(minimal.study, "sponsor", globals.errors_and_logging)
  result = document._translate_references('<usdm:macro id="section" name="inclusion"/>', False)
  assert str(result) == '<usdm:macro id="section" name="inclusion"></usdm:macro>'

def test_translate_rererences_2(mocker, globals, minimal, factory):
  criteria = create_criteria(factory, minimal)
  document = Document(minimal.study, "sponsor", globals.errors_and_logging)
  result = document._translate_references('<usdm:ref id="EligibilityCriterion_1" klass="EligibilityCriterion" attribute="text"/>', False)
  assert str(result) == 'Only perform at baseline'

def test_content_to_html_1(mocker, globals, minimal, factory):
  criteria = create_criteria(factory, minimal)
  item =  factory.item(NarrativeContentItem, {'name': "NCI1", 'text': '<usdm:macro id="section" name="inclusion"/>', 'childIds': []})
  content = factory.item(NarrativeContent, {'name': "NC1", 'sectionNumber': '1.1.1', 'displaySectionNumber': True, 'sectionTitle': 'Section Title', 'displaySectionTitle': True, 'contentItemId': item.id, 'childIds': []})
  minimal.study.versions[0].narrativeContentItems = [item]
  minimal.study.documentedBy[0].versions[0].contents = [content]
  doc = Doc()
  document = Document(minimal.study, "sponsor", globals.errors_and_logging)
  document._content_to_html(content, doc)
  result = doc.getvalue()
  expected = '<div class=""><h3 id="section-1.1.1">1.1.1 Section Title</h3><usdm:macro id="section" name="inclusion"></usdm:macro></div>'
  assert result == expected




from usdm_excel.ncit import NCIt


def test_code(mocker, globals):
    mock_id = mocker.patch("usdm_excel.id_manager.IdManager.build_id")
    mock_id.side_effect=['Code_1']
    item = NCIt(globals)
    code = item.code(code="CODE", decode="DECODE")
    assert code.id == "Code_1"
    assert code.code == "CODE"
    assert code.codeSystem == "NCI Thesaurus"
    assert code.codeSystemVersion == "24.01e"
    assert code.decode == "DECODE"


import pytest
import pandas as pd
from usdm_excel.study_design_epoch_sheet.study_design_epoch_sheet import StudyDesignEpochSheet
from usdm_model.code import Code


xfail = pytest.mark.xfail

def test_create(mocker, globals):
  mock_id = mocker.patch("usdm_excel.id_manager.IdManager.build_id")
  mock_id.side_effect=['EpochId_1', 'EpochId_2', 'EpochId_3']
  expected_1 = Code(id='Code1', code='code', codeSystem='codesys', codeSystemVersion='3', decode="label1")
  expected_2 = Code(id='Code2', code='code', codeSystem='codesys', codeSystemVersion='3', decode="label2")
  expected_3 = Code(id='Code3', code='code', codeSystem='codesys', codeSystemVersion='3', decode="label3")
  mock_code = mocker.patch("usdm_excel.cdisc_ct.CDISCCT.code_for_attribute")
  mock_code.side_effect=[expected_1, expected_2, expected_3]
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [['Epoch 1', 'Epoch One', 'C12345'], ['Epoch 2', 'Epoch Two', 'C12345'], ['Epoch 3', 'Epoch Three', 'C12345']]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['studyEpochName', 'studyEpochDescription', 'studyEpochType'])
  epochs = StudyDesignEpochSheet("", globals)
  assert len(epochs.items) == 3
  assert epochs.items[0].id == 'EpochId_1'
  assert epochs.items[0].name == 'Epoch 1'
  assert epochs.items[1].id == 'EpochId_2'
  assert epochs.items[1].description == 'Epoch Two'
  assert epochs.items[2].id == 'EpochId_3'
  assert epochs.items[2].type == expected_3
  
def test_create_with_label(mocker, globals):
  mock_id = mocker.patch("usdm_excel.id_manager.IdManager.build_id")
  mock_id.side_effect=['EpochId_1', 'EpochId_2', 'EpochId_3']
  expected_1 = Code(id='Code1', code='code', codeSystem='codesys', codeSystemVersion='3', decode="label1")
  expected_2 = Code(id='Code2', code='code', codeSystem='codesys', codeSystemVersion='3', decode="label2")
  expected_3 = Code(id='Code3', code='code', codeSystem='codesys', codeSystemVersion='3', decode="label3")
  mock_code = mocker.patch("usdm_excel.cdisc_ct.CDISCCT.code_for_attribute")
  mock_code.side_effect=[expected_1, expected_2, expected_3]
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [['Epoch 1', 'Epoch One', '1', 'C12345'], ['Epoch 2', 'Epoch Two', '2', 'C12345'], ['Epoch 3', 'Epoch Three', '', 'C12345']]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['name', 'description', 'label', 'studyEpochType'])
  epochs = StudyDesignEpochSheet("", globals)
  assert len(epochs.items) == 3
  assert epochs.items[0].id == 'EpochId_1'
  assert epochs.items[0].name == 'Epoch 1'
  assert epochs.items[0].label == '1'
  assert epochs.items[1].id == 'EpochId_2'
  assert epochs.items[1].description == 'Epoch Two'
  assert epochs.items[2].id == 'EpochId_3'
  assert epochs.items[2].type == expected_3
  assert epochs.items[2].label == ''
  
def test_create_empty(mocker, globals):
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = []
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['studyEpochName', 'studyEpochDescription', 'studyEpochType'])
  epochs = StudyDesignEpochSheet("", globals)
  assert len(epochs.items) == 0

def test_read_cell_by_name_error(mocker, globals):
  mock_error = mocker.patch("usdm_excel.errors_and_logging.errors.Errors.add")
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [['Epoch 1', 'Epoch One']]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['studyEpochName', 'studyEpochDescription'])
  epochs = StudyDesignEpochSheet("", globals)
  mock_error.assert_called()
  assert mock_error.call_args[0][0] == "studyDesignEpochs"
  assert mock_error.call_args[0][1] == None
  assert mock_error.call_args[0][2] == None
  assert mock_error.call_args[0][3] == "Exception. Error [Failed to detect column(s) 'studyEpochType, type' in sheet] while reading sheet 'studyDesignEpochs'. See log for additional details."
  



from typing import List, Literal
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc, ApiBaseModelWithId

class ParameterMap(ApiBaseModelWithId):
  tag: str
  reference: str
  instanceType: Literal['ParameterMap']

class SyntaxTemplateDictionary(ApiBaseModelWithIdNameLabelAndDesc):
  parameterMaps: List[ParameterMap] = [] # Allow for empty list, not in API
  instanceType: Literal['SyntaxTemplateDictionary']



from typing import List, Literal, Union
from .activity import Activity
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .alias_code import AliasCode
from .biomedical_concept import BiomedicalConcept
from .biomedical_concept_category import BiomedicalConceptCategory
from .biomedical_concept_surrogate import BiomedicalConceptSurrogate
from .code import Code
from .encounter import Encounter
from .study_cell import StudyCell
from .indication import Indication
from .study_intervention import StudyIntervention
from .study_arm import StudyArm
from .study_epoch import StudyEpoch
from .study_element import StudyElement
from .population_definition import StudyDesignPopulation
from .objective import Objective
from .schedule_timeline import ScheduleTimeline
from .estimand import Estimand
from .syntax_template_dictionary import SyntaxTemplateDictionary
from .condition import Condition
from .comment_annotation import CommentAnnotation

class StudyDesign(ApiBaseModelWithIdNameLabelAndDesc):
  trialIntentTypes: List[Code] = []
  trialTypes: List[Code] = []
  therapeuticAreas: List[Code] = []
  characteristics: List[Code] = []
  interventionModel: Code
  encounters: List[Encounter] = []
  activities: List[Activity] = []
  biomedicalConcepts: List[BiomedicalConcept] = []
  bcCategories: List[BiomedicalConceptCategory] = []
  bcSurrogates: List[BiomedicalConceptSurrogate] = []
  arms: List[StudyArm]
  studyCells: List[StudyCell]
  blindingSchema: Union[AliasCode, None] = None
  rationale: str
  epochs: List[StudyEpoch]
  elements: List[StudyElement] = []
  estimands: List[Estimand] = []
  indications: List[Indication] = []
  studyInterventions: List[StudyIntervention] = []
  objectives: List[Objective] = []
  population: Union[StudyDesignPopulation, None] = None
  scheduleTimelines: List[ScheduleTimeline] = []
  documentVersionId: Union[str, None] = None
  dictionaries: List[SyntaxTemplateDictionary] = []
  conditions: List[Condition] = []
  notes: List[CommentAnnotation] = []
  instanceType: Literal['StudyDesign']

  def main_timeline(self):
    return next((item for item in self.scheduleTimelines if item.mainTimeline), None)


from usdm_excel.globals import Globals
from usdm_model.code import Code

class CodeBase():

  def __init__(self, globals: Globals):
    self._globals = globals
    self._id_manager = globals.id_manager
    self._cross_references = globals.cross_references

  def _build(self, code, system, version, decode):
    id = self._id_manager.build_id(Code)
    instance = Code(id=id, code=code, codeSystem=system, codeSystemVersion=version, decode=decode)
    self._cross_references.add(instance.id, instance)
    return instance
 


from typing import List, Literal
from .api_base_model import ApiBaseModelWithId

class StudyCell(ApiBaseModelWithId):
  armId: str
  epochId: str
  elementIds: List[str] = []
  instanceType: Literal['StudyCell']



import os
import docraptor
from yattag import Doc
from usdm_model.study import Study
from usdm_model.narrative_content import NarrativeContent, NarrativeContentItem
from usdm_db.cross_reference import CrossReference
from usdm_db.errors_and_logging.errors_and_logging import ErrorsAndLogging
from usdm_db.document.utility import get_soup

class Document():

  class LogicError(Exception):
    pass

  def __init__(self, study: Study, template_name: str, errors_and_logging: ErrorsAndLogging):
    self._study = study
    self._errors_and_logging = errors_and_logging
    self._errors_and_logging.debug(f"Initialising document for template '{template_name}'")
    self._cross_ref = CrossReference(study, self._errors_and_logging)
    try:
      self._study_version = self._study.versions[0]
      self._study_design = self._study_version.studyDesigns[0]
      self._document = self._study.document_by_template_name(template_name)
      self._errors_and_logging.debug(f"Document '{self._document}'")
      self._document_version = self._document.versions[0]
      title = self._study_version.get_title('Official Study Title')
      self._doc_title = title.text if title else '[Document Title]'
      self._chapters = []
      self._modal_count = 1
    except Exception as e:
      self._errors_and_logging.exception(f"Failed to initialise for document creation", e)

  def to_pdf(self, test=True):
    doc_api = docraptor.DocApi()
    doc_api.api_client.configuration.username = os.getenv('DOCRAPTOR_API_KEY')
    document_content = self.to_html()
    try:
      response = doc_api.create_doc({
        'test': test,  # test documents are free but watermarked
        #'test': False,  # Non-watermarked documents, but limited number allowed.
        'document_type': 'pdf',
        'document_content': document_content,
        # 'document_url': 'https://docraptor.com/examples/invoice.html',
        # 'javascript': True,
        # 'prince_options': # {
        #    'media': 'print', # @media 'screen' or 'print' CSS
        #    'baseurl': 'https://yoursite.com', # the base URL for any relative URLs
        # },
      })
      binary_formatted_response = bytearray(response)
      return binary_formatted_response
    except docraptor.rest.ApiException as e:
      self._errors_and_logging.exception(f"Exception raised generating PDF content. See logs for more details", e)
      return None
    except Exception as e:
      self._errors_and_logging.exception(f"Exception raised generating PDF content. See logs for more details", e)
      return None

  def to_html(self, highlight=False):
    try:
      self._modal_count = 1
      self.chapters = []
      #root = self._document_version.contents[0]
      doc = Doc()
      doc.asis('<!DOCTYPE html>')
      nc_list = self._document_version.narrative_content_in_order()
      for content in nc_list:
        if self._is_level_1_doc_section(content.sectionNumber):
          self.chapters.append(f'<a href="#section-{content.sectionNumber}"></a>')
      with doc.tag('html'):
        with doc.tag('head'):
          with doc.tag('meta', **{'charset': "utf-8"}):
            pass
          with doc.tag('meta', **{"name": "viewport", "content": "width=device-width", "initial-scale": "1"}):
            pass
          with doc.tag('style'):
            doc.asis(self._style())     
          
          # Google fonts
          attributes = {
            'href': "https://fonts.googleapis.com/css2?family=Open+Sans:wght@400;600;800&display=swap",
            'rel': "stylesheet"
          }
          with doc.tag('link', **attributes):
            pass

          # Bootstrap Icons
          attributes = {
            'href': "https://cdn.jsdelivr.net/npm/bootstrap-icons@1.9.1/font/bootstrap-icons.css",
            'rel': "stylesheet",
            'type': "text/css"
          }
          with doc.tag('link', **attributes):
            pass
          
          # Bootstrap CSS
          attributes = {
            'href': "https://cdn.jsdelivr.net/npm/bootstrap@5.3.2/dist/css/bootstrap.min.css",
            'rel': "stylesheet",
            'integrity': "sha384-T3c6CoIi6uLrA9TneNEoa7RxnatzjcDSCmG1MXxSR1GAsXEV/Dwwykc2MPK8M2HN",
            'crossorigin': "anonymous"
          }
          with doc.tag('link', **attributes):
            pass

          # Bootstrap JS
          attributes = {
            'src': "https://cdn.jsdelivr.net/npm/bootstrap@5.3.2/dist/js/bootstrap.bundle.min.js",
            'integrity': "sha384-C6RzsynM9kWDrMNeT87bh95OGNyZPhcTNXj1NW7RuBCsyN/o0jlpcV8Qyq46cDfL",
            'crossorigin': "anonymous"
          }
          with doc.tag('script', **attributes):
            pass
        
        with doc.tag('body'):
          doc.asis(self._title_page())
          #nc_list = self._document_version.narrative_content_in_order()
          for content in nc_list:
            self._content_to_html(content, doc, highlight)
      return doc.getvalue()
    except Exception as e:
      self._errors_and_logging.exception(f"Exception raised generating HTML content. See logs for more details", e)
      return None

  def _content_to_html(self, content: NarrativeContent, doc, highlight: bool=False) -> None:
    level = self._get_level(content.sectionNumber)
    #print(f"LEVEL: L={level} C={content}")
    klass = "page" if level == 1 else ""
    heading_id = f"section-{content.sectionNumber}"
    if (level == 1 and self._is_first_section(content.sectionNumber)):
      with doc.tag('div', klass=klass):
        doc.asis(self._table_of_contents())    
    with doc.tag('div', klass=klass):
      if (self._is_level_1_doc_section(content.sectionNumber)) or (level > 1):
        with doc.tag(f'h{level}', id=heading_id):
          doc.asis(f"{content.sectionNumber} {content.sectionTitle}")
      content_item = self._cross_ref.get(NarrativeContentItem, content.contentItemId)
      if content_item:
        doc.asis(str(self._translate_references(content_item.text, highlight)))
      else:
        self._errors_and_logging.error(f"Missing content item while processing content for section '{content.sectionNumber}', '{content.sectionTitle}' while generating the HTML document")

  def _get_level(self, section_number):
    if section_number.lower().startswith("appendix"):
      result = 1
    else:
      text = section_number[:-1] if section_number.endswith('.') else section_number
      result = len(text.split('.'))
    return result
      
  def _is_doc_section(self, section_number):
    try:
      sn = int(section_number)
      return True if sn > 0 else False
    except:
      return True

  def _is_level_1_doc_section(self, section_number):
    level = self._get_level(section_number)
    return level == 1 and self._is_doc_section(section_number)
  
  def _is_first_section(self, section_number):
    try:
      sn = int(section_number)
      return True if sn == 1 else False
    except:
      return False

  def _translate_references(self, content_text, highlight=False):
    #print(f"HIGHLIGHT TR: {highlight}")
    soup = get_soup(content_text, self._errors_and_logging)
    for ref in soup(['usdm:ref']):
      try:
        attributes = ref.attrs
        instance = self._cross_ref.get(attributes['klass'], attributes['id'])
        #print(f"REF: {instance}")
        value = self._resolve_instance(instance, attributes['attribute'], highlight)
        translated_text = self._translate_references(value, highlight)
        self._replace_and_highlight(soup, ref, translated_text, highlight)
      except Exception as e:
        self._errors_and_logging.exception(f"Exception raised while attempting to translate reference '{attributes}' while generating the HTML document, see the logs for more info", e)
        self._replace_and_highlight(soup, ref, 'Missing content: exception', highlight)
    self._errors_and_logging.debug(f"Translate references from {content_text} => {get_soup(str(soup), self._errors_and_logging)}")
    return get_soup(str(soup), self._errors_and_logging)

  def _resolve_instance(self, instance, attribute, highlight=False):
    #print(f"RI: {instance}, {attribute} {highlight}")
    dictionary = self._get_dictionary(instance)
    value = str(getattr(instance, attribute))
    soup = get_soup(value, self._errors_and_logging)
    for ref in soup(['usdm:tag']):
      try:
        attributes = ref.attrs
        if dictionary:
          entry = next((item for item in dictionary.parameterMaps if item.tag == attributes['name']), None)
          self._replace_and_highlight(soup, ref, get_soup(entry.reference, self._errors_and_logging), highlight)
        else:
          self._errors_and_logging.error(f"Missing dictionary while attempting to resolve reference '{attributes}' while generating the HTML document")
          self._replace_and_highlight(soup, ref, 'Missing content: missing dictionary', highlight)
      except Exception as e:
        self._errors_and_logging.exception(f"Failed to resolve reference '{attributes} while generating the HTML document", e)
        self._replace_and_highlight(soup, ref, 'Missing content: exception', highlight)
    return str(soup)

  def _get_dictionary(self, instance):
    try:
      return self._cross_ref.get('SyntaxTemplateDictionary', instance.dictionaryId)
    except:
      return None  

  def _replace_and_highlight(self, soup, ref, text, highlight):
    if highlight:
      self._wrap_in_span_and_modal(soup, ref, text)
    else:
      ref.replace_with(text)

  def _wrap_in_span_and_modal(self, soup, ref, text):
    id = f"usdmContent{self._modal_count}"
    span = soup.new_tag('span', attrs={'class': "usdm-highlight"})
    span.append(get_soup(self._link(id), self._errors_and_logging))
    span.append(text)
    x = ref.replace_with(span)
    span.append(get_soup(self._modal(ref, id), self._errors_and_logging))
    self._modal_count += 1

  def _link(self, id):
    return f"""
      <a class="link-dark usdm-highlight-link" style="font-size: 12px;" data-bs-toggle="modal" data-bs-target="#{id}">
        <i class="ps-1 pe-1 bi bi-info-circle"/>
      </a>
    """

  def _modal(self, ref, id):
    body = [f"<b>'{k}':</b> '{v}'" for k,v in ref.attrs.items()]
    return f"""
      <div class="modal fade" id="{id}" tabindex="-1">
        <div class="modal-dialog">
          <div class="modal-content">
            <div class="modal-header">
              <h5 class="modal-title">Included using '{ref.name}'</h5>
              <button type="button" class="btn-close" data-bs-dismiss="modal" aria-label="Close"></button>
            </div>
            <div class="modal-body">
              Attributes: {(', ').join(body)}
            </div>
          </div>
        </div>
      </div>
    """

  def _table_of_contents(self):
    return f"""
      <div id="toc-page" class="page">
        <div id="table-of-contents">
          {''.join(self.chapters)}
        </div>
        <div align="center">
          <br/>
          <p class="usdm-warning">Note:</p>
          <p class="usdm-warning">The table of contents is auto generated upon PDF production and only includes first level sections.</p>
        </div>
        <div id="header-and-footer">
          <span id="page-number"></span>
        </div>
      </div>
    """

  def _title_page(self):
    return f"""
      <div id="title-page" class="page">
        <div class="container">
          <div class="row">
            <div class="col-md-8 offset-md-2 text-center">
              <h1>{self._doc_title}</h1>
            </div>
          </div>
          <div class="row mt-5">
            <div class="col-md-8 offset-md-2 text-center">
              <p class="usdm-warning">Note:</p>
              <p class="usdm-warning">This document is generated from content held within the Unified Studies Definitions Model. It is for test purposes only.</p>
            </div>
          </div>
        </div>
        <div id="header-and-footer">
          <span id="page-number"></span>
        </div>
      </div>
    """

  def _style(self):
    return """
      /* Create a running element */
      #header-and-footer {
        position: running(header-and-footer);
        text-align: right;
      }

      /* Add that running element to the top and bottom of every page */
      @page {
        @top {
          content: element(header-and-footer);
        }
        @bottom {
          content: element(header-and-footer);
        }
      }

      /* Add a page number */
      #page-number {
        content: "Page " counter(page);
      }

      @page { size: A4 }

      /* Create a title page with a full-bleed background and no header */
      /* Create a soa page with landscape orientation */
      #title-page {
        page: title-page;
      }
      .soa-page {
        page: soa-page
      }

      @page title-page {
        @top {
          content: "";
        }
      }

      @page soa-page {
        size: A4 landscape
      }

      #title-page h1 {
        padding: 200px 0 40px 0;
        font-size: 30px;
      }

      /* Dynamically create a table of contents with leaders */
      #table-of-contents a {
        content: target-content(attr(href)) leader('.') target-counter(attr(href), page);
        color: #000000;
        text-decoration: none;
        display: block;
        padding-top: 5px;
      }

      /* Float the footnote to a footnotes area on the page */
      .footnote {
        float: footnote;
        font-size: small;
      }

      .page {
        page-break-after: always;
      }

      body {
        counter-reset: chapter;
        font-family: 'Times New Roman';
        color: #000000;
      }

      .usdm-warning {
        font-style: italic;
        color: red;
      }

      .soa-body-text {
        font-size: 10px;
        text-align: center
      }

      .soa-activity-text {
        font-size: 10px;
        text-align: left
      }

      .soa-footnote-text {
        font-size: 12px;
        text-align: left
      }

      .usdm-highlight {
        background-color: LightGray !important;
      }

      .usdm-highlight p {
        background-color: LightGray;
      }

      p .usdm-highlight {
        background-color: LightGray;
      }

      .usdm-highlight-link {
        text-decoration:none;
      }

    """  


from usdm_model.study_design import StudyDesign
from usdm_model.study_epoch import StudyEpoch
from usdm_model.study_cell import StudyCell
from usdm_model.study_arm import StudyArm
from usdm_model.population_definition import *
from usdm_model.study import Study
from usdm_model.study_version import StudyVersion
from usdm_model.study_title import StudyTitle
from usdm_model.study_definition_document import StudyDefinitionDocument
from usdm_model.study_definition_document_version import StudyDefinitionDocumentVersion
from usdm_model.identifier import StudyIdentifier
from usdm_model.organization import Organization
from usdm_model.address import Address
from usdm_model.governance_date import GovernanceDate
from usdm_model.geographic_scope import GeographicScope
from usdm_model.subject_enrollment import SubjectEnrollment
from usdm_model.study_amendment import StudyAmendment
from usdm_model.study_amendment_reason import StudyAmendmentReason
from usdm_model.quantity import Quantity
from usdm_excel.globals import Globals
from tests.test_factory import Factory

class MinimalStudy():

  def __init__(self, globals: Globals):
    globals.id_manager.clear()
    factory = Factory(globals)
    feedback_reason = factory.item(StudyAmendmentReason, {'code': factory.cdisc_code("C99904x3", "IRB/IEC Feedback")})
    other_reason = factory.item(StudyAmendmentReason, {'code': factory.cdisc_code("C17649", "Other"), "otherReason": "Fix typographical errors"})
    subjects = factory.item(Quantity, {'value': 10.0})
    enrollment = factory.item(SubjectEnrollment, {'name': 'ENROLL1', 'type': factory.cdisc_code("C68846", "Global"), 'quantity': subjects})
    geo_scope = factory.item(GeographicScope, {'type': factory.cdisc_code("C68846", "Global")})
    amendment = factory.item(StudyAmendment,{"number": "1", "summary": "Updated inclusion criteria", "substantialImpact": True,
                                             "primaryReason": feedback_reason, "secondaryReasons": [other_reason], "enrollments": [enrollment], 
                                             "geographicScopes": [geo_scope]}) 
    global_scope = factory.item(GeographicScope, {'type': factory.cdisc_code("C68846", "Global")})
    europe_code = factory.alias_code(factory.geo_code('150', 'Europe'), [])
    europe_scope = factory.item(GeographicScope, {'type': factory.cdisc_code("C41129", "Region"), "code": europe_code})
    study_approval_date = factory.item(GovernanceDate, {"name": "D_APPROVE", "label": "Design Approval", "description": "Design approval date", 
                                                  "type": factory.cdisc_code("C132352", "Sponsor Approval Date"), "dateValue": "2006-06-01", "geographicScopes": [global_scope]})
    doc_approval_date = factory.item(GovernanceDate, {"name": "D_APPROVE", "label": "Design Approval", "description": "Design approval date", 
                                             "type": factory.cdisc_code("C99903x1", "Sponsor Approval Date"), "dateValue": "2006-06-01", "geographicScopes": [europe_scope]})
    phase_code = factory.cdisc_code('C12345', 'Phase Code')
    alias_phase = factory.alias_code(phase_code, [])
    self.population = factory.item(StudyDesignPopulation, {'name': 'POP1', 'label': '', 'description': '', 'includesHealthySubjects': True, 'criteria': []})
    cell = factory.item(StudyCell, {'armId': "X", 'epochId': "Y"})
    arm = factory.item(StudyArm, {'name': "Arm1", 'type': factory.cdisc_dummy(), 'dataOriginDescription': 'xxx', 'dataOriginType': factory.cdisc_dummy()})
    epoch = factory.item(StudyEpoch, {'name': 'EP1', 'label': 'Epoch A', 'description': '', 'type': factory.cdisc_code('C22222', 'Epoch Code')})
    study_title = factory.item(StudyTitle, {'text': 'Title', 'type': factory.cdisc_code('C44444', 'Official Study Title')})
    study_short_title = factory.item(StudyTitle, {'text': 'Short Title', 'type': factory.cdisc_code('C33333', 'Brief Study Title')})
    study_acronym = factory.item(StudyTitle, {'text': 'ACRONYM', 'type': factory.cdisc_code('C33333', 'Study Acronym')})
    self.study_definition_document_version = factory.item(StudyDefinitionDocumentVersion, {'version': '1', 'status': factory.cdisc_dummy(), 'dateValues': [doc_approval_date]})
    self.study_definition_document = factory.item(StudyDefinitionDocument, {'name': 'PD1', 'label': 'Protocol Document', 'description': '', 
      'language': factory.english(), 'type': factory.cdisc_code('C70817', 'Protocol'), 'templateName': "Sponsor", 
      'versions': [self.study_definition_document_version]})
    self.study_design = factory.item(StudyDesign, {'name': 'Study Design', 'label': '', 'description': '', 
      'rationale': 'Study Design Rationale', 'interventionModel': factory.cdisc_dummy(), 'arms': [arm], 'studyCells': [cell], 
      'epochs': [epoch], 'population': self.population})
    address = factory.item(Address, {'line': 'line 1', 'city': 'City', 'district': 'District', 'state': 'State', 'postalCode': '12345', 'country': factory.code("UKK", "UKK_decode")})
    organization_1 = factory.item(Organization, {'name': 'Sponsor', 'type': factory.cdisc_code("C70793", "sponsor"), 'identifier': "123456789", 'identifierScheme': "DUNS", 
                                                 'legalAddress': address}) 
    identifier = factory.item(StudyIdentifier, {'text': 'SPONSOR-1234', 'scopeId': organization_1.id})
    organization_2 = factory.item(Organization, {'name': 'Sponsor', 'type': factory.cdisc_code("C188863", "reg 1"), 
                                                 'identifier': "REG 1", 'identifierScheme': "DUNS", 'legalAddress': address}) 
    reg_1_identifier = factory.item(StudyIdentifier, {'text': 'REG 111111', 'scopeId': organization_2.id})
    organization_3 = factory.item(Organization, {'name': 'Sponsor', 'type': factory.cdisc_code("C93453", "reg 2"), 
                                                 'identifier': "REG 2", 'identifierScheme': "DUNS", 'legalAddress': address}) 
    reg_2_identifier = factory.item(StudyIdentifier, {'text': 'REG 222222', 'scopeId': organization_3.id})
    self.study_version = factory.item(StudyVersion, {'versionIdentifier': '1', 'rationale': 'Study version rationale', 'titles': [study_title, study_short_title, study_acronym], 
                                                     'studyDesigns': [self.study_design], 
                                                     'documentVersionId': self.study_definition_document_version.id, 'studyIdentifiers': [identifier, reg_1_identifier, reg_2_identifier], 
                                                     'studyPhase': alias_phase, 'dateValues': [study_approval_date], 'amendments': [amendment], 'organizations': [organization_1, organization_2, organization_3]})
    self.study = factory.item(Study, {'id': None, 'name': 'Study', 'label': '', 'description': '', 'versions': [self.study_version], 'documentedBy': [self.study_definition_document]}) 
    #print(f"MINIMAL: {self.study.to_json()}")



from usdm_excel.document.utility import get_soup

def test_normal(mocker, globals, factory):
  globals.errors_and_logging.errors().clear()
  bs = factory.base_sheet(mocker)
  assert globals.errors_and_logging.errors().count() == 0
  result = get_soup("<p>Hello</p>", bs)
  expected = '<p>Hello</p>'
  assert str(result) == expected
  assert globals.errors_and_logging.errors().count() == 0

def test_warning(mocker, globals, factory):
  mock_error = mocker.patch("usdm_excel.errors_and_logging.errors_and_logging.ErrorsAndLogging.debug")
  globals.errors_and_logging.errors().clear()
  bs = factory.base_sheet(mocker)
  assert globals.errors_and_logging.errors().count() == 0
  result = get_soup("input/output", bs)
  expected = 'input/output'
  assert str(result) == expected
  mock_error.assert_called()
  assert mock_error.call_args[0][0] == "Warning raised within Soup package, processing 'input/output'\nMessage returned 'The input looks more like a filename than markup. You may want to open this file and pass the filehandle into Beautiful Soup.'"
  assert mock_error.call_args[0][1] == ''




from usdm_excel.iso_8601_duration import ISO8601Duration
from usdm_excel.range_type import RangeType
from usdm_excel.globals import Globals

class WindowType():

  def __init__(self, timing_info: str, globals: Globals):
    self.upper = None
    self.lower = None
    self.errors = []
    self.label = timing_info.strip() if timing_info else ""
    if self.label:
      range = RangeType(self.label, globals)
      if not range.errors:
        self.lower = self._set_encoded(range.lower, range.units)
        self.upper = self._set_encoded(range.upper, range.units)     
      else:
        self.errors = range.errors

  def _set_encoded(self, value, units):
    for char in ['+', '-']:
      if char in value:
        value = value.replace(char, "")
    value = ISO8601Duration().encode(value, units)
    return value



import traceback
from usdm_excel.base_sheet import BaseSheet
from usdm_model.study_amendment import StudyAmendment
from usdm_model.study_amendment_reason import StudyAmendmentReason
from usdm_model.subject_enrollment import SubjectEnrollment
from usdm_model.quantity import Quantity
from usdm_excel.cdisc_ct import CDISCCT
from usdm_excel.iso_3166 import ISO3166
from usdm_excel.alias import Alias
from usdm_excel.quantity_type import QuantityType
from usdm_model.governance_date import GovernanceDate
from usdm_excel.globals import Globals

class StudyAmendmentSheet(BaseSheet):

  SHEET_NAME = 'studyAmendments'
  
  def __init__(self, file_path: str, globals: Globals):
    try:
      self.items = []
      super().__init__(file_path=file_path, globals=globals, sheet_name=self.SHEET_NAME, optional=True)
      if self.success:
        for index, row in self.sheet.iterrows():
          secondaries = []
          number = self.read_cell_by_name(index, 'number')
          date_name = self.read_cell_by_name(index, 'date', must_be_present=False)
          date = self.globals.cross_references.get(GovernanceDate, date_name)
          summary = self.read_cell_by_name(index, 'summary')
          #substantial = self.read_boolean_cell_by_name(index, 'substantialImpact')
          notes = self.read_cell_multiple_by_name(index, 'notes', must_be_present=False)
          primary_reason = self._read_primary_reason_cell(index)
          primary = self._amendment_reason(primary_reason)
          secondary_reasons = self._read_secondary_reason_cell(index)
          for reason in secondary_reasons:
            amendment_reason = self._amendment_reason(reason)        
            if amendment_reason:
              secondaries.append(amendment_reason)
          enrollments = self._read_enrollment_cell(index)
          scopes = self.read_geographic_scopes_cell_by_name(index, 'geographicScope')
          params = {
            'number': number,
            'summary': summary,
            #'substantialImpact': substantial,
            'primaryReason': primary,
            'secondaryReasons': secondaries,
            'enrollments': enrollments,
            'geographicScopes': scopes,
            'dateValues': [date] if date else []
          }
          item = self.create_object(StudyAmendment, params)
          if item:
            self.items.append(item)
            self.globals.cross_references.add(item.number, item)
            self.add_notes(item, notes)
        self.items.sort(key=lambda d: int(d.number))
        self.previous_link(self.items, 'previousId')
        
    except Exception as e:
      self._sheet_exception(e)

  def _amendment_reason(self, reason):
    item = self.create_object(StudyAmendmentReason, {'code': reason['code'], 'otherReason': reason['other']})
    return item
    
  def _read_enrollment_cell(self, row_index):
    result = []
    col_index = self._get_column_index('enrollment')
    value = self.read_cell(row_index, col_index)
    if value.strip() == '':
      self._error(row_index, col_index, "Empty cell detected where enrollment values expected")
    else:
      for item in self._state_split(value):
        key_value = self._key_value(item, row_index, col_index)
        if key_value[0] == "COHORT":
          pass
        elif key_value[0] == "SITE":
          pass
        elif key_value[0] == "GLOBAL":
          quantity = self._get_quantity(key_value[1])
          scope = self._scope('Global', None)
          result.append(self._enrollment(quantity, scope=scope))
        elif key_value[0] == "REGION": 
          code, quantity = self._country_region_quantity(key_value[1], 'Region', row_index, col_index)
          if code:
            scope = self._scope('Region', code)
            result.append(self._enrollment(quantity, scope=scope))
        elif key_value[0] == "COUNTRY": 
          code, quantity = self._country_region_quantity(key_value[1], 'Country', row_index, col_index)
          if code:
            scope = self._scope('Country', code)
            result.append(self._enrollment(quantity, scope=scope))
    return result

  def _enrollment(self, quantity, **kwargs):
    applies_to = None
    applies_to_id = None
    if 'scope' in kwargs:
      applies_to = kwargs['scope']
    if 'cohort' in kwargs:
      applies_to_id = kwargs['cohort']
    if 'site' in kwargs:
      applies_to_id = kwargs['site']
    return self.create_object(SubjectEnrollment, {'name': "XXX", 'quantity': quantity, 'appliesTo': applies_to, 'appliesToId': applies_to_id})

  def _read_secondary_reason_cell(self, row_index):
    results = []
    col_index = self._get_column_index('secondaryReasons')
    value = self.read_cell(row_index, col_index)
    if not value.strip():
      return results
    parts = value.strip().split(',')
    for part in parts:
      result = self._extract_reason(part, row_index, col_index)
      if result:
        results.append(result)
    return results
      
  def _read_primary_reason_cell(self, row_index):
    col_index = self._get_column_index('primaryReason')
    value = self.read_cell(row_index, col_index)
    return self._extract_reason(value, row_index, col_index)
  
  def _extract_reason(self, value, row_index, col_index):
    if value.strip() == "":
      self._error(row_index, col_index, "Empty cell detected where CDISC CT value expected.")
      return None
    elif value.strip().upper().startswith('OTHER'):
      text = value.strip()
      parts = text.split("=")
      if len(parts) == 2:
        return {'code': CDISCCT(self.globals).code_for_attribute('StudyAmendmentReason', 'code', 'Other'), 'other': parts[1].strip()}
      else:
        self._error(row_index, col_index, f"Failed to decode reason data {text}, no '=' detected")
    else:
      code = CDISCCT(self.globals).code_for_attribute('StudyAmendmentReason', 'code', value)
      if code is None:
        self._error(row_index, col_index, f"CDISC CT not found for value '{value}'.")
      return {'code': code, 'other': None}


from typing import List, Literal, Union
from .api_base_model import ApiBaseModelWithIdNameAndLabel
from .code import Code
from .address import Address
from .study_site import StudySite

class Organization(ApiBaseModelWithIdNameAndLabel):
  type: Code
  identifierScheme: str
  identifier: str
  legalAddress: Union[Address, None] = None
  managedSites: List[StudySite] = []
  instanceType: Literal['Organization']



from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from typing import Literal

class TransitionRule(ApiBaseModelWithIdNameLabelAndDesc):
  text: str
  instanceType: Literal['TransitionRule']



import logging
import traceback
import warnings
from usdm_excel.errors_and_logging.errors import Errors

class ErrorsAndLogging():

  WARNING = Errors.WARNING
  ERROR = Errors.ERROR
  DEBUG = Errors.DEBUG
  INFO = Errors.INFO

  def __init__(self):
    self._logger = logging.getLogger(__name__)
    self._logger.addHandler(logging.NullHandler())
    self._errors = Errors()

  def errors(self):
    return self._errors
  
  def debug(self, message: str, sheet: str=None, row: int=None, column: int=None):
    self._logger.debug(self._format(message, sheet, row, column))

  def info(self, message: str, sheet: str=None, row: int=None, column: int=None):
    self._logger.info(self._format(message, sheet, row, column))

  def exception(self, message: str, e: Exception, sheet: str=None, row: int=None, column: int=None):
    self._errors.add(sheet, row, column, f"Exception. {message}. See log for additional details.", self._errors.ERROR)
    self._logger.error(f"Exception '{e}' raised\n\n{self._format(message, sheet, row, column)}\n\n{traceback.format_exc()}")

  def warning(self, message: str, sheet: str=None, row: int=None, column: int=None):
    self._errors.add(sheet, row, column, message, self._errors.WARNING)
    self._logger.warning(self._format(message, sheet, row, column))

  def error(self, message: str, sheet: str=None, row: int=None, column: int=None):
    self._errors.add(sheet, row, column, message, self._errors.ERROR)
    self._logger.error(self._format(message, sheet, row, column))

  def deprecated(self, message):
    warnings.warn(message, DeprecationWarning)    
    self._logger.warning(self._format(message, None, None, None))

  def _format(self, message, sheet, row, column):
    if not sheet:
      return f"{message}"
    elif row == None:
      return f"In sheet {sheet}: {message}"
    else:
      return f"In sheet {sheet} at [{row},{column}]: {message}"



import os
import base64
from usdm_excel.document.template_m11 import TemplateM11
from usdm_excel.document.template_plain import TemplatePlain
from usdm_excel.document.elements import Elements
from usdm_excel.base_sheet import BaseSheet
from usdm_model.abbreviation import Abbreviation
from usdm_model.study_version import StudyVersion
from usdm_model.study_definition_document_version import StudyDefinitionDocumentVersion
from usdm_model.activity import Activity
from usdm_model.biomedical_concept import BiomedicalConcept
from usdm_model.biomedical_concept_surrogate import BiomedicalConceptSurrogate
from usdm_excel.document.utility import get_soup

class Macros():

  def __init__(self, parent: BaseSheet, study_version: StudyVersion, document_version: StudyDefinitionDocumentVersion):
    self._parent = parent
    self._study_version = study_version
    self._study_design = self._study_version.studyDesigns[0]
    self._document_version = document_version

    self._elements = Elements(parent, study_version, document_version)
    self._m11 = TemplateM11(parent, study_version, document_version)
    self._plain = TemplatePlain(parent, study_version, document_version)
    self._template_map = {'m11': self._m11, 'plain': self._plain}
  
  def resolve(self, content_text: str) -> str:
    soup = get_soup(content_text, self._parent)
    for ref in soup(['usdm:macro']):
      try:
        attributes = ref.attrs
        method = f"_{attributes['id'].lower()}"
        if self._valid_method(method):
          result = getattr(self, method)(attributes, soup, ref)
        else:
          self._parent._general_error(f"Failed to resolve document macro '{attributes}', invalid method name {method}")
          ref.replace_with('Missing content: invalid method name')
      except Exception as e:
        self._parent._general_exception(f"Failed to resolve document macro '{attributes}'", e)
        ref.replace_with('Missing content: exception')
    return str(soup)

  def _xref(self, attributes: dict, soup, ref) -> None:
    instance, attribute = self._parent.globals.cross_references.get_by_path(attributes['klass'], attributes['name'], attributes['attribute'])
    ref_tag = soup.new_tag("usdm:ref")
    ref_tag.attrs = {'klass': instance.__class__.__name__, 'id': instance.id, 'attribute': attribute}
    ref.replace_with(ref_tag)
      
  def _abbreviations(self, attributes: dict, soup, ref) -> None:
    items = [x.strip() for x in attributes['items'].split(',')]
    separator = attributes['separator'] if 'separator' in attributes else ','
    first = True
    for item in items:
      instance = self._parent.globals.cross_references.get(Abbreviation, item)
      if instance:
        if not first:
          ref.insert_before(f'{separator} ')
        ref.insert_before(self._new_ref_tag(soup, 'Abbreviation', instance.id, 'abbreviatedText'))
        ref.insert_before(' = ')
        ref.insert_before(self._new_ref_tag(soup, 'Abbreviation', instance.id, 'expandedText'))
        first = False
      else:
        self._parent._general_error(f"Failed to find abbreviation '{item}'")
        ref.insert_after(f"Missing abbreviation: failed to find abbreviation '{item}'")
    ref.extract()

  def _new_ref_tag(self, soup, klass: str, id: str, attribute: str):
    tag = soup.new_tag("usdm:ref")
    tag.attrs = {'klass': klass, 'id': id, 'attribute': attribute}
    return tag
        
  def _bc(self, attributes: dict, soup, ref) -> None:
    bc_name = attributes['name'].upper().strip()
    activity_name = attributes['activity'].strip()
    activity = self._parent.globals.cross_references.get(Activity, activity_name)
    if activity:
      bc = None
      for collection in [{'klass': BiomedicalConcept, 'ids': activity.biomedicalConceptIds, 'synonyms': True}, 
                         {'klass': BiomedicalConceptSurrogate, 'ids': activity.bcSurrogateIds, 'synonyms': False}]:
        for id in collection['ids']:
          next_bc = self._parent.globals.cross_references.get_by_id(collection['klass'], id)
          if next_bc:
            if bc_name == next_bc.name.upper() or bc_name in [x.upper() for x in next_bc.synonyms]:
              bc = next_bc
              break
      if bc:      
        ref_tag = soup.new_tag("usdm:ref")
        ref_tag.attrs = {'klass': bc.__class__.__name__, 'id': bc.id, 'attribute': 'label'}
        ref.replace_with(ref_tag)
      else:
        self._parent._general_error(f"Failed to find BC name '{bc_name}' in activity '{activity_name}'")
        ref.replace_with('Missing BC: failed to find BC in activity')
    else:
      self._parent._general_error(f"Failed to find activity '{activity_name}'")
      ref.replace_with(f"Missing activity: failed to find activity '{activity_name}'")

  def _image(self, attributes: dict, soup, ref) -> None:
    type = attributes['type']
    data = self._encode_image(attributes['file'])
    if data:
      img_tag = soup.new_tag("img")
      img_tag.attrs['src'] = f"data:image/{type};base64,{data.decode('ascii')}"
      img_tag.attrs['alt'] = "Alt text"
      ref.replace_with(img_tag)
    else:
      name = attributes['file']
      self._parent._general_error(f"Failed to insert image '{name}'")
      ref.replace_with(f"Missing image: failed to insert image '{name}'")

  def _element(self, attributes, soup, ref) -> None:
    method = attributes['name'].lower()
    if self._elements.valid_method(method):
      text = getattr(self._elements, method)()
      ref.replace_with(get_soup(text, self._parent))
    else:
      self._parent._general_error(f"Failed to translate element method name '{method}' in '{attributes}', invalid method")
      ref.replace_with('Missing content: invalid method name')

  def _section(self, attributes, soup, ref) -> None:
    method = attributes['name'].lower()
    template = attributes['template'] if 'template' in attributes else 'plain' 
    instance = self._resolve_template(template)
    if instance.valid_method(method):
      text = getattr(instance, method)(attributes)
      ref.replace_with(get_soup(text, self._parent))
    else:
      self._parent._general_error(f"Failed to translate section method name '{method}' in '{attributes}', invalid method")
      ref.replace_with('Missing content: invalid method name')       

  def _note(self, attributes, soup, ref) -> None:
    text = f"""
      <div class="col-md-8 offset-md-2 text-center">
        <p class="usdm-warning">Note:</p>
        <p class="usdm-warning">{attributes['text']}</p>
      </div>
    """
    ref.replace_with(get_soup(text, self._parent))

  def _valid_method(self, name):
    return name in ['_xref', '_image', '_element', '_section', '_note', '_bc', '_abbreviations']
  
  def _resolve_template(self, template: str) -> object:
    try:
      return self._template_map[template.lower()]
    except:
      self._parent._general_error(f"Failed to map template '{template}', using plain template")
      return self._plain

  def _encode_image(self, filename: str) -> bytes:
    try:
      with open(os.path.join(self._parent.dir_path, filename), "rb") as image_file:
        data = base64.b64encode(image_file.read())
      return data
    except Exception as e:
      self._parent._general_exception(f"Failed to open / read image file '{filename}', ignored", e)
      return None
      